import logging
from nabu.utils import Progress
from nabu.stitching.slurm_utils import split_stitching_configuration_to_slurm_job
from .cli_configs import StitchingConfig
from ..pipeline.config import parse_nabu_config_file
from nabu.stitching.z_stitching import (
    z_stitching,
    StitchingPostProcAggregation,
)
from nabu.stitching.config import dict_to_config_obj
from .utils import parse_params_values

try:
    from sluurp.executor import submit
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True

_logger = logging.getLogger(__name__)


def main():
    args = parse_params_values(
        StitchingConfig,
        parser_description="Run stitching from a configuration file. Configuration can be obtain from `stitching-config`",
    )
    logging.basicConfig(level=args["loglevel"].upper())

    conf_dict = parse_nabu_config_file(args["input-file"], allow_no_value=True)

    stitching_config = dict_to_config_obj(conf_dict)
    stitching_config.settle_inputs()
    if args["only_create_master_file"]:
        # option to ease creation of the master in the following cases:
        # * user has submitted all the job but has been quicked out of the cluster
        # * only a few slurm job for some random version (cluster update...) and user want to retriger only those job and process the aggragation only. On those cases no need to redo it all.
        tomo_objs = []
        for _, sub_config in split_stitching_configuration_to_slurm_job(stitching_config, yield_configuration=True):
            tomo_objs.append(sub_config.get_output_object().get_identifier().to_str())

        post_processing = StitchingPostProcAggregation(
            existing_objs_ids=tomo_objs,
            stitching_config=stitching_config,
        )
        post_processing.process()

    elif stitching_config.slurm_config.partition in (None, ""):
        # case 1: run locally
        _logger.info(f"run stitching locally with {stitching_config}")

        progress = Progress("z-stitching")
        progress.set_name("initialize z-stitching")
        progress.set_advancement(0)
        z_stitching(stitching_config, progress=progress)
    else:
        if not has_sluurp:
            raise ImportError(
                "sluurp not install. Please install it to distribute stitching on slurm (pip install sluurm)"
            )
        # case 2: run on slurm
        # note: to speed up we could do shift research on pre processing and run it only once (if manual of course). Here it will be run for all part
        _logger.info(f"will distribute stitching")

        futures = {}
        # 2.1 launch jobs
        for i_job, (job, sub_config) in enumerate(
            split_stitching_configuration_to_slurm_job(stitching_config, yield_configuration=True)
        ):
            _logger.info(f"submit job nb {i_job}: handles {sub_config.slices}")
            output_volume = sub_config.get_output_object().get_identifier().to_str()
            futures[output_volume] = submit(job, timeout=999999)

        # 2.2 wait for future to be done and concatenate the result
        post_processing = StitchingPostProcAggregation(
            futures=futures,
            stitching_config=stitching_config,
        )
        post_processing.process()

    exit(0)


if __name__ == "__main__":
    main()



import numpy as np
from scipy import ndimage
import logging

logger = logging.getLogger(__name__)


def find_center_vo(tomo, ind=None, smin=-50, smax=50, srad=6, step=0.25,
                   ratio=0.5, drop=20):
    """
    Find rotation axis location using Nghia Vo's method. :cite:`Vo:14`.

    Parameters
    ----------
    tomo : ndarray
        3D tomographic data or a 2D sinogram.
    ind : int, optional
        Index of the slice to be used for reconstruction.
    smin, smax : int, optional
        Coarse search radius. Reference to the horizontal center of
        the sinogram.
    srad : float, optional
        Fine search radius.
    step : float, optional
        Step of fine searching.
    ratio : float, optional
        The ratio between the FOV of the camera and the size of object.
        It's used to generate the mask.
    drop : int, optional
        Drop lines around vertical center of the mask.
    ncore : int, optional
        Number of cores that will be assigned to jobs.

    Returns
    -------
    float
        Rotation axis location.
    """
    if tomo.ndim == 2:
        tomo = np.expand_dims(tomo, 1)
        ind = 0
    (depth, height, width) = tomo.shape
    if ind is None:
        ind = height // 2
        if height > 10:
            # Averaging sinograms to improve SNR
            _tomo = np.mean(tomo[:, ind - 5:ind + 5, :], axis=1)
        else:
            _tomo = tomo[:, ind, :]
    else:
        _tomo = tomo[:, ind, :]

    # Denoising
    # There's a critical reason to use different window sizes
    # between coarse and fine search.
    _tomo_cs = ndimage.filters.gaussian_filter(_tomo, (3, 1), mode='reflect')
    _tomo_fs = ndimage.filters.gaussian_filter(_tomo, (2, 2), mode='reflect')

    # Coarse and fine searches for finding the rotation center.
    if _tomo.shape[0] * _tomo.shape[1] > 4e6:  # If data is large (>2kx2k)
        _tomo_coarse = _tomo_cs[::4, ::4] # TODO proper downsampling
        init_cen = _search_coarse(
            _tomo_coarse, smin / 4.0, smax / 4.0, ratio, drop)
        fine_cen = _search_fine(_tomo_fs, srad, step,
                                init_cen * 4.0, ratio, drop)
    else:
        init_cen = _search_coarse(_tomo_cs, smin, smax, ratio, drop)
        fine_cen = _search_fine(_tomo_fs, srad, step,
                                init_cen, ratio, drop)
    logger.debug('Rotation center search finished: %i', fine_cen)
    return fine_cen




def _calculate_metric(shift_col, sino1, sino2, sino3, mask):
    """
    Metric calculation.
    """
    shift_col = 1.0 * np.squeeze(shift_col)
    if np.abs(shift_col - np.floor(shift_col)) == 0.0:
        shift_col = int(shift_col)
        sino_shift = np.roll(sino2, shift_col, axis=1)
        if shift_col >= 0:
            sino_shift[:, :shift_col] = sino3[:, :shift_col]
        else:
            sino_shift[:, shift_col:] = sino3[:, shift_col:]
        mat = np.vstack((sino1, sino_shift))
    else:
        sino_shift = ndimage.interpolation.shift(
            sino2, (0, shift_col), order=3, prefilter=True)
        if shift_col >= 0:
            shift_int = int(np.ceil(shift_col))
            sino_shift[:, :shift_int] = sino3[:, :shift_int]
        else:
            shift_int = int(np.floor(shift_col))
            sino_shift[:, shift_int:] = sino3[:, shift_int:]
        mat = np.vstack((sino1, sino_shift))
    metric = np.mean(
        np.abs(np.fft.fftshift(np.fft.fft2(mat))) * mask)
    return np.asarray([metric], dtype=np.float32)


def _search_coarse(sino, smin, smax, ratio, drop):
    """
    Coarse search for finding the rotation center.
    """
    (nrow, ncol) = sino.shape
    cen_fliplr = (ncol - 1.0) / 2.0
    smin = np.int16(np.clip(smin + cen_fliplr, 0, ncol - 1) - cen_fliplr)
    smax = np.int16(np.clip(smax + cen_fliplr, 0, ncol - 1) - cen_fliplr)
    start_cor = ncol // 2 + smin
    stop_cor = ncol // 2 + smax
    flip_sino = np.fliplr(sino)
    comp_sino = np.flipud(sino)  # Used to avoid local minima
    list_cor = np.arange(start_cor, stop_cor + 0.5, 0.5)
    list_metric = np.zeros(len(list_cor), dtype=np.float32)
    mask = _create_mask(2 * nrow, ncol, 0.5 * ratio * ncol, drop)
    list_shift = 2.0 * (list_cor - cen_fliplr)

    #
    shifts = np.float32(list_shift)
    list_metric = []
    for shift in shifts:
        list_metric.append(
            _calculate_metric(shift, sino, flip_sino, comp_sino, mask)
        )
    #
    minpos = np.argmin(list_metric)
    if minpos == 0:
        logger.debug('WARNING!!!Global minimum is out of searching range')
        logger.debug('Please extend smin: %i', smin)
    if minpos == len(list_metric) - 1:
        logger.debug('WARNING!!!Global minimum is out of searching range')
        logger.debug('Please extend smax: %i', smax)
    cor = list_cor[minpos]
    return cor


def _search_fine(sino, srad, step, init_cen, ratio, drop):
    """
    Fine search for finding the rotation center.
    """
    (nrow, ncol) = sino.shape
    cen_fliplr = (ncol - 1.0) / 2.0
    srad = np.clip(np.abs(srad), 1.0, ncol / 4.0)
    step = np.clip(np.abs(step), 0.1, srad)
    init_cen = np.clip(init_cen, srad, ncol - srad - 1)
    list_cor = init_cen + np.arange(-srad, srad + step, step)
    flip_sino = np.fliplr(sino)
    comp_sino = np.flipud(sino)
    mask = _create_mask(2 * nrow, ncol, 0.5 * ratio * ncol, drop)
    list_shift = 2.0 * (list_cor - cen_fliplr)

    #
    shifts = np.float32(list_shift)
    list_metric = []
    for shift in shifts:
        list_metric.append(
            _calculate_metric(shift, sino, flip_sino, comp_sino, mask)
        )
    #
    cor = list_cor[np.argmin(list_metric)]
    return cor



def _create_mask(nrow, ncol, radius, drop):
    """
    Make a binary mask to select coefficients outside the double-wedge region.
    Eq.(3) in https://doi.org/10.1364/OE.22.019078

    Parameters
    ----------
    nrow : int
        Image height.
    ncol : int
        Image width.
    radius: int
        Radius of an object, in pixel unit.
    drop : int
        Drop lines around vertical center of the mask.

    Returns
    -------
        2D binary mask.
    """
    du = 1.0 / ncol
    dv = (nrow - 1.0) / (nrow * 2.0 * np.pi)
    cen_row = np.int16(np.ceil(nrow / 2.0) - 1)
    cen_col = np.int16(np.ceil(ncol / 2.0) - 1)
    drop = min(drop, np.int16(np.ceil(0.05 * nrow)))
    mask = np.zeros((nrow, ncol), dtype='float32')
    for i in range(nrow):
        pos = np.int16(np.ceil(((i - cen_row) * dv / radius) / du))
        (pos1, pos2) = np.clip(np.sort(
            (-pos + cen_col, pos + cen_col)), 0, ncol - 1)
        mask[i, pos1:pos2 + 1] = 1.0
    mask[cen_row - drop:cen_row + drop + 1, :] = 0.0
    mask[:, cen_col - 1:cen_col + 2] = 0.0
    return mask






import os
import pip
from subprocess import check_output, getoutput
from os.path import dirname
import fnmatch


def get_nvidia_driver_version():
    cmd = 'dpkg -l | grep nvidia-cuda-toolkit | tr -s " " | cut -d " " -f 3'
    return getoutput(cmd)

def get_os(sep=" "):
    cmd1 = "lsb_release -is"
    cmd2 = "lsb_release -rs"
    return sep.join([getoutput(cmd) for cmd in [cmd1, cmd2]])

def get_machine_name():
    cmd = "hostname"
    return getoutput(cmd)

def get_machine_type():
    cmd = "uname -m"
    return getoutput(cmd)


def get_site_packages():
    return dirname(dirname(pip.__file__))


def is_conda_activated():
    return bool(int(os.environ.get("CONDA_SHLVL", "0")))


def get_env_name(verbose=False):
    def verbose_print(msg):
        if verbose:
            print(msg)

    hostname = get_machine_name()
    slurm_partition = os.environ.get("SLURM_JOB_PARTITION", None)

    if slurm_partition is not None:
        verbose_print("Detected slurm partition %s" % slurm_partition)
        env = slurm_partitions_mappings.get(slurm_partition, None)
        if env is not None:
            verbose_print("Using environment %s" % env)
            return env
        else:
            verbose_print("No environment found for this partition")

    verbose_print("Trying individual nodes rules")
    for node_wildcard, env in nodes_mappings.items():
        if fnmatch.filter([hostname], node_wildcard):
            verbose_print("Match with %s -> %s" % (node_wildcard, env))
            return env

    verbose_print("Could not find any environment for %s" % hostname)
    return None








import numpy as np
import matplotlib.pyplot as plt
from nabu.estimation.tilt import CameraTilt
import fabio

def BOLD(x):
    pass


def nabu_align():
    with fabio.open('/tmp/bl/scan0001/edgetwinmic_0000.h5') as f:
        flat = f.data
    with fabio.open('/tmp/bl/scan0002/edgetwinmic_0000.h5') as f:
        dark = f.data
    with fabio.open('/tmp/bl/scan0003/edgetwinmic_0000.h5') as f:
        proj0 = f.data
    with fabio.open('/tmp/bl/scan0003/edgetwinmic_0001.h5') as f:
        proj180 = f.data
    radio0 = (proj0.astype(float) - dark.astype(float)) / (
            flat.astype(float) - dark.astype(float)
    )
    radio180 = (proj180.astype(float) - dark.astype(float)) / (
            flat.astype(float) - dark.astype(float)
    )
    radio180_flip = np.fliplr(radio180.copy())
    tilt_calc = CameraTilt()
    # enable calculation results plotting
    tilt_calc.verbose = True
    plt.subplot(1, 2, 1)
    plt.imshow(radio0)
    plt.subplot(1, 2, 2)
    plt.imshow(radio180_flip)
    plt.show()
    pixel_cor, camera_tilt = tilt_calc.compute_angle(radio0, radio180_flip)#, method="fft-polar")
    print("CameraTilt: pixel_cor = %f" % pixel_cor)
    pos_cor = 0.0003632 * pixel_cor

    print(BOLD(f"Lateral alignment position correction in mm: {-pos_cor:>10.5f}"))
    print(
        BOLD(
            f"Camera tilt in deg : {camera_tilt:35.5f}\n"
        )
    )

    input("Press any key")
    plt.close("all")

if __name__ == "__main__":
    nabu_align()



from datetime import datetime
import numpy as np
import matplotlib.pyplot as plt

steps_to_measure = [
    "Reading data",
    "Applying flat-field",
    "Applying double flat-field",
    "Applying CCD corrections",
    "Performing phase retrieval",
    "Performing unsharp mask",
    "Taking logarithm",
    "Building sinograms",
    "Removing rings on sinograms",
    "Reconstruction",
    "Computing histogram",
    "Saving data"
]

def parse_reconstruction(lines, separator=" - "):
    def extract_timestamp(line):
        timestamp = line.split(separator)[0]
        return datetime.strptime(timestamp, "%d-%m-%Y %H:%M:%S")

    def extract_current_step(line):
        return line.split(separator)[-1]

    current_step = extract_current_step(lines[0])
    t1 = extract_timestamp(lines[0])

    res = {}
    for line in lines[1:]:
        line = line.strip()
        if len(line.split(separator)) == 1:
            continue
        timestamp = line.strip().split(separator)[0]
        t2 = datetime.strptime(timestamp, "%d-%m-%Y %H:%M:%S")

        res.setdefault(current_step, [])
        res[current_step].append((t2 - t1).seconds)

        t1 = t2
        current_step = extract_current_step(line)

    return res


def parse_logfile(fname, separator=" - "):
    with open(fname, "r") as f:
        lines = f.readlines()

    start_text = "Going to reconstruct slices"
    end_text = "Merging reconstructions to"

    rec_log_bounds = []
    for i, line in enumerate(lines):
        if start_text in line:
            start_line = i
        if end_text in line:
            rec_log_bounds.append((start_line, i))


    results = {}
    for bounds in rec_log_bounds:
        start, end = bounds
        results[start] = {}
        res = parse_reconstruction(lines[start:end], separator=separator)
        for step in steps_to_measure:
            if step in res:
                results[start][step] = {
                    "n": len(res[step]),
                    "mean": np.mean(res[step]),
                    "min": np.min(res[step]),
                    "max": np.max(res[step]),
                    "std": np.std(res[step])
                }
    return results



# https://matplotlib.org/stable/gallery/lines_bars_and_markers/horizontal_barchart_distribution.html#sphx-glr-gallery-lines-bars-and-markers-horizontal-barchart-distribution-py
def survey(results, category_names, colormap="tab10"):
    """
    Parameters
    ----------
    results : dict
        A mapping from question labels to a list of answers per category.
        It is assumed all lists contain the same number of entries and that
        it matches the length of *category_names*.
    category_names : list of str
        The category labels.
    """
    labels = list(results.keys())
    data = np.array(list(results.values()))
    data_cum = data.cumsum(axis=1)
    category_colors = plt.get_cmap(colormap)(
        # np.linspace(0.15, 0.85, data.shape[1])
        np.linspace(0., 1, data.shape[1])
    )

    fig, ax = plt.subplots(figsize=(9.2, 5))
    ax.invert_yaxis()
    ax.xaxis.set_visible(False)
    ax.set_xlim(0, np.sum(data, axis=1).max())

    for i, (colname, color) in enumerate(zip(category_names, category_colors)):
        widths = data[:, i]
        starts = data_cum[:, i] - widths
        rects = ax.barh(labels, widths, left=starts, height=0.5,
                        label=colname, color=color)

        r, g, b, _ = color
        text_color = 'white' if r * g * b < 0.5 else 'darkgrey'
        ax.bar_label(rects, label_type='center', color=text_color)
    ax.legend(ncol=len(category_names), bbox_to_anchor=(0, 1),
              loc='lower left', fontsize='small')

    return fig, ax



def plot_results(results, colormap="tab10"):
    # This assumes that the processing steps are always the same between runs!
    plot_values = {}
    for i, result in enumerate(results.values()):
        categories = result.keys()
        plot_values["Run %d" % (i + 1)] = [v["mean"] for v in result.values()]
    fig, ax = survey(plot_values, categories, colormap=colormap)
    plt.show()






import os
from time import time, sleep
from multiprocessing import Process, Queue, current_process
path = os.path
from integrator.integrator import StackIntegrator
from integrator.app.utils import get_distributed_integrator_params

os.environ["OMP_NUM_THREADS"] = "4"

def worker(ai_config, dataset, out_dir, task_queue, out_queue):

    os.environ["OMP_NUM_THREADS"] = "4"
    S = StackIntegrator(ai_config)
    sleep(5)
    S.set_new_dataset(dataset, out_dir, 200)
    print("OK!")

    for task in iter(task_queue.get, 'STOP'):
        i_min, i_max = task
        res = S.process_stack(i_min, i_max)
        out_queue.put(res)


def worker_scan(ai_config, dataset, out_dir):
    S = StackIntegrator(ai_config)
    sleep(5)
    S.set_new_dataset(dataset, out_dir, 200)
    print("OK!")

    for i in range(int(25e3)):
        i_min, i_max = (200 * i, 200 * (i+1))
        res = S.process_stack(i_min, i_max)


def integrate_dataset(ai_config, dataset, out_dir, n_proc):


    tasks = []
    for i in range(int(25e3 / 200)):
        tasks.append((200 * i, 200 * (i+1)))


    # Create queues
    task_queue = Queue()
    done_queue = Queue()

    # Start worker processes
    for i in range(n_proc):
        Process(target=worker, args=(ai_config, dataset, out_dir, task_queue, done_queue)).start()


    input("Start ?")

    # Submit tasks
    for task in tasks:
        task_queue.put(task)




    # Get and print results
    print('Unordered results:')
    for i in range(len(tasks)):
        print('\t', done_queue.get())

    # Tell child processes to stop
    for i in range(n_proc):
        task_queue.put('STOP')



conf, ai_config, cluster, datasets, output_files = get_distributed_integrator_params("/home/esrf/paleo/tmp/xrdct_sessions/ihma109/test.conf")
integrate_dataset(ai_config, datasets[-19], path.dirname(output_files[-19]), 8)


from math import floor, ceil
import numpy as np
from scipy.ndimage import convolve1d

def shift_bilinear2(img, shift, axis=1):
    """
    Shift an image with bilinear interpolation,
    For a translation, the difference between non-interpolated and interpolated coordinates is constant.
    Thus, the bilinear weighting it can be implemented as a convolution operator.
    """
    if img.ndim != 2:
        raise ValueError("Expected 2D image")
    axis = axis % 2
    res = np.zeros_like(img)
    # Extract the integer and fractional parts of the shift
    shift0 = shift
    shift = abs(shift)
    shift_i, shift_f = int(shift), shift - int(shift)
    shift1, shift2 = shift - floor(shift), ceil(shift) - shift

    # Integer shift is simply a slicing operation
    # Fractional shift is a convolution operation
    res_area = slice(shift_i, None)
    img_area = slice(0, -shift_i)
    if shift < 0: # negative shift
        res_area, img_area = img_area, res_area

    slice_all = slice(None, None)
    if axis == 1:
        res_area = (slice_all, res_area)
        img_subset = img[slice_all, img_area]
    else:
        res_area = (res_area, slice_all)
        img_subset = img[img_area, slice_all]

    if shift_f < 1e-4:
        res[res_area] = img_subset
    else:
        res[res_area] = convolve1d(img_subset, [shift1, shift2], axis=axis, mode="reflect")
    return res





from time import time
from threading import Thread
import numpy as np
from tomoscan.io import HDF5File

class Processor:
    def __init__(self, size=8000):
        self.A = np.random.rand(size, size).astype("f")
        self._cnt = 0


    def process(self):
        print("Processing %d..." % self._cnt)
        t0 = time()
        _ = self.A.T.dot(self.A)
        el = time() - t0
        print("process OK in %.2f s" % el)
        self._cnt += 1




class Reader:
    def __init__(self, fname, h5path, start_z=None, end_z=None):
        self.fname = fname
        self.h5path = h5path
        self.set_subregion(start_z, end_z)


    def set_subregion(self, start_z, end_z):
        self.start_z = start_z
        self.end_z = end_z


    def read(self):
        print("Reading (%d, %d)..." % (self.start_z, self.end_z))
        t0 = time()
        with HDF5File(self.fname, "r") as f:
            d = f[self.h5path][:, self.start_z:self.end_z, :]
        el = time() - t0
        print(
            "Read (%d, %d) in %.2f s"
            % (self.start_z, self.end_z, el)
        )
        return d


if __name__ == "__main__":
    fname = "/data/scisofttmp/tomo_datasets/rock/sample_0002_1_1.nx"
    h5path = "entry0000/instrument/detector/data"
    R = Reader(fname, h5path, start_z=0, end_z=100)

    P = Processor()

    reader_thread = Thread(target=R.read)
    reader_thread.start()

    P.process()



def interleaved_read_proc(fname, h5path, n_steps, chunk_size=100):

    R = Reader(fname, h5path, start_z=0, end_z=chunk_size)
    P = Processor(size=5000)

    def read_in_thread(start_z, end_z):
        R.set_subregion(start_z, end_z)
        reader_thread = Thread(target=R.read)
        reader_thread.start()
        return reader_thread

    def wait_for_available_data(reader_thread):
        print("[Processor] waiting for available data...")
        reader_thread.join()
        print("[Processor] OK!")

    t0 = time()

    reader_thread = read_in_thread(0, chunk_size)

    for i in range(1, n_steps+1):
        wait_for_available_data(reader_thread)
        # transfer R -> P
        reader_thread = read_in_thread(i * chunk_size, (i + 1) * chunk_size)
        P.process()

    el = time() - t0
    print("Total time: %.3fs" % el)



"""
Independently:
  - read: 12.3s
  - process: 19.3

In parallel:
  - read: 14.3
  - process: 20.7



[Read][T][Proces][Read][T][Process][Read][T][Process][Read][T][Process]

[Read][T][Process][T][Process][T][Process][T][Process]
         [Read]......[Read]......[Read]......[Read]




[Read][T][Process][Save][Read][T][Process][Save][Read][T][Process][Save][Read][T][Process][Save]


[Read][T][Process]......[T][Process][T][Process][T][Process]
         [Read]......[Read]......[Read]......[Read]
                  [Save]











"""


import numpy as np
from nabu.preproc.phase import PaganinPhaseRetrieval


def get_decay(curve, cutoff=1e3, vmax=None):
    if vmax is None:
        vmax = curve.max()
    return np.argmax(curve < vmax / cutoff)



def compute_paganin_margin(shape, **pag_kwargs):
    """

    Parameters
    -----------
    shape: tuple
        Detector shape in the form (n_z, n_x)
    """
    P = PaganinPhaseRetrieval(shape, **pag_kwargs)

    ifft_func = np.fft.irfft2 if P.use_R2C else np.fft.ifft2
    conv_kernel = ifft_func(P.paganin_filter)

    vmax = conv_kernel[0,0]

    v_margin = get_decay(conv_kernel[:, 0], cutoff=1e3, vmax=vmax)
    h_margin = get_decay(conv_kernel[0, :], cutoff=1e3, vmax=vmax)
    # If the Paganin filter is very narrow, then the corresponding convolution
    # kernel is constant, and np.argmax() gives 0 (when it should give the max value)
    if v_margin == 0:
        v_margin = shape[0]
    if h_margin == 0:
        h_margin = shape[1]

    return v_margin, h_margin





import numpy as np
from nabu.resources.dataset_analyzer import analyze_dataset
from nabu.io.reader import ChunkReader
from nabu.preproc.ccd import FlatField
from nabu.preproc.double_flat_field import DoubleFlatField
from nabu.preproc.ccd import CCDCorrection
from nabu.preproc.phase import PaganinPhaseRetrieval
from nabu.preproc.sinogram import SinoProcessing

from nabu.reconstruction.fbp import Backprojector
from spire.utils import ims


di = analyze_dataset("/data/visitor/md1252/bm05/HR/tube02/HA-800_2.25um_FO-20.122ULZ-OLZ-OLP_008_")

C = ChunkReader(di.projections, sub_region=(None, None, 1024-50, 1024+50), convert_float=True)
C.load_files()
radios = C.files_data

F2 =FlatField(radios.shape, di.flats, di.darks, radios_indices=sorted(di.projections.keys()), sub_region=C.sub_region)
F2.normalize_radios(radios)

DFF = DoubleFlatField(radios.shape, input_is_mlog=False)
DFF.apply_double_flatfield(radios)

ccd_corr = CCDCorrection(radios.shape)
ccd_corr.median_clip_correction(radios, inplace=True)

P2 = PaganinPhaseRetrieval(radios.shape[1:], distance=di.distance*1e1, energy=di.energy, pixel_size=2.25, delta_beta=100)
for i in range(radios.shape[0]):
    radios[i] = P2.apply_filter(radios[i])

np.log(radios, out=radios)
radios *= -1
S = SinoProcessing(radios_shape=radios.shape, rot_center=1831.5, halftomo=True)
o = S.radios_to_sinos(radios)

B = Backprojector(o.shape[1:], extra_options={"padding_mode": "edges"})
ims(B.fbp(o[50]), cmap="gray")







# Taken from PyHST/Cspace/CCspace.c (line 6066 @ f826cfc0de7be7ee185bb6ba38f7fdb0d0ec2a8a)
def straighten_sino(Sino, output=None):
    if output is None:
        output = Sino
    Nr, Nc = Sino.shape
    J = np.arange(Nc)
    x = 2.* (J + 0.5 - Nc/2)/Nc
    sum0 = Nc
    f2 = (3.0*x*x-1.0)
    sum1 = (x**2).sum()
    sum2 = (f2**2).sum()

    for i in range(Nr):
        ff0 = Sino[i, :].sum()
        ff1 = (x * Sino[i, :]).sum()
        ff2 = (f2*Sino[i, :]).sum()

        output[i, :] = Sino[i, :] - ff0/sum0 + ff1*x/sum1 + ff2*f2/sum2

    return output








import numpy as np
from silx.io import get_data
from nabu.preproc.ccd import Log
from nabu.preproc.flatfield import FlatFieldDataUrls
from nabu.estimation.cor_sino import SinoCor

class SinoRed:
    """
    Class and method to prepare sinogram and calculate COR in HA
    The pseudo sinogram is built with shrinked radios taken every theta degres

        ovs: oversampling horizontaly the radios. Paul use 4. Here 1 is used because float subpixel shift is used instead
        theta: Take a radio every theta degres and its correspondant at 180 degrees.
        yred: resize vertically the radio to yred lines
    """
    def __init__(self, dataset_info, side="right", oversampling=4, theta=10, subsampling_y=10, take_log=True):
                
        self.take_log = take_log
        self.side = side
        self.ovs = oversampling
        self.theta = theta
        self.yred_fact = subsampling_y
        self.dataset_info = dataset_info

        self.nproj = self.dataset_info.n_angles

        self.dproj = round(self.nproj/360*self.theta)
        self.proj_stop = round(self.nproj/(2*self.dproj))+1
        self.sx, self.sy = self.dataset_info.radio_dims
        self.rcor_abs = round(self.sx/2.)
        self.cor_acc = round(self.sx/2.)

        # in this mode, the sinogram will be composed by a succession of 360/dtheta radios
        # 180/dtheta should be integer. sy/syred_fact should be integer
        # The radios themseleves are compressed verticaly by a factor of yred_fact.
        # in the original Paul's algorithm, the radio are horizontally oversampled by 4 to obtain
        # the subpixel precision. Here we prefor to do a float schift on 1/10 pixels to obtain this
        # subprecision.
        
        self.nprojred = 2*self.proj_stop
        self.yred = round(self.sy/self.yred_fact)

        # initialize sinograms and radios arrays
        self.sino = np.zeros((int(self.nprojred * self.yred), self.sx))

        self.projs_indices = np.arange(self.proj_stop)*self.dproj
        self.projs_absolute_indices = sorted(self.dataset_info.projections.keys())

        self.flatfield = FlatFieldDataUrls(
            (len(self.projs_indices), self.sy, self.sx),
            proc.dataset_info.flats,
            proc.dataset_info.darks,
            radios_indices=[self.projs_absolute_indices[i] for i in self.projs_indices],
            dtype=np.float64
        )
        self.mlog = Log(
            (1, ) + self.flatfield.shape,
            clip_min=1e-6,
            clip_max=10.
        )

    def get_radio(self, image_num):
        radio_dataset_idx = self.projs_absolute_indices[image_num]
        data_url = self.dataset_info.projections[radio_dataset_idx]
        radio = get_data(data_url).astype(np.float64)
        self.flatfield.normalize_single_radio(radio, radio_dataset_idx, dtype=radio.dtype)
        if self.take_log:
            self.mlog.take_logarithm(radio)
        return radio
    
    def get_sino(self):
        """
        Build sinogram (composite image) from the radio files
        """
        nradios = np.arange(self.proj_stop)*self.dproj
        np2 = round(self.nproj/2)
        ns2 = round(self.sino.shape[0]/2)
        # loop on all the projections
        irad = 0
        for npi in self.projs_indices:
            radio1 = np.resize(self.get_radio(npi), (self.yred, self.sx))
            radio2 = np.resize(self.get_radio(npi+np2), (self.yred, self.sx))
            
            self.sino[irad:irad+self.yred,:] = radio1
            self.sino[irad+ns2:irad+ns2+self.yred,:] = radio2
            
            irad = irad + self.yred
                
        self.sino[np.isnan(self.sino)] = 0.0001 # ?
        return self.sino



from os import getpid
from time import sleep
from multiprocessing import Process
import sys

from silx.gui import qt
from nabu.resources.processconfig import ProcessConfig

from nabu.app.fullfield_cuda import CudaFullFieldPipeline


"""
Idéalement, on ne crée qu'une seule fois un objet "CudaFullFieldPipeline".
Ca engendrerait moins de problèmes de création/suppression de contextes.
Pour ce faire :
  - Créer cet objet dans OuterThread, et partager avec InnerThread.
    Par sûr que ça marche avec les QThread
  - Créer un InnerThread persistant (donc un objet CudaFullFieldPipeline persistant)
    qui agit comme un serveur : OuterThread envoie des messages de reconstruction
    (ex. queue). Là encore, pas sûr qu'on puisse utiliser ça avec QThread

Pour tomwer, le nombre de slices à reconstruire est variable !
Donc dans tous les cas, certainement on détruit InnerThread
"""


# This thread always use the same FullFieldPipeline instance
class InnerThread1(qt.QThread):

    def __init__(self, proc, subregion, **kwargs):
        super().__init__(**kwargs)
        self.pipeline = CudaFullFieldPipeline(
            proc, subregion, cuda_options={"cleanup_at_exit": False}
        )

    def run(self, subregion):
        print("In Inner QThread v1")
        self.pipeline.process_chunk(sub_region=subregion)
        print("End Inner QThread v1")

    def __del__(self):
        print("Inner QThread v1 being destroyed. Deleting cuda context")
        self.pipeline.ctx.pop()


# This thread creates a new FullFieldPipeline object each time
class InnerThread2(qt.QThread):

    def __init__(self, fname, subregion, **kwargs):
        super().__init__(**kwargs)
        self.proc = ProcessConfig(fname)
        self.subregion = subregion
        self.pipeline = CudaFullFieldPipeline(
            self.proc, self.subregion, cuda_options={"cleanup_at_exit": False}
        )

    def run(self):
        print("In Inner QThread v2")
        # necessary if CudaFullFieldPipeline was created in __init__
        self.pipeline.ctx.push()
        #
        self.pipeline.process_chunk()
        print("End Inner QThread v2")
        self.pipeline.ctx.pop()
        del self.pipeline
        sleep(1)

    # ~ def __del__(self):
        # ~ self.pipeline.ctx.pop()
        # ~ print("Destroying Inner QThread v2 - deleting cuda context")


# Thread calling InnerThread1
class OuterThread(qt.QThread):

    def __init__(self, fname, **kwargs):
        super().__init__(**kwargs)
        self.fname = fname

    def run(self):
        fname = self.fname
        print("In Outer QThread")
        I = InnerThread2(fname, (100, 150))
        I.start()
        I.wait()
        print("End Outer Qthread")
        sleep(1)
        del I
        sleep(1)
        # ~ print("Launching a second time")
        # ~ I = InnerThread2(fname, (150, 200))
        # ~ I.start()
        # ~ I.wait()
        # ~ print("End second time")

        # ~ print("Launching a third time")
        # ~ I = InnerThread2(fname, (200, 250))
        # ~ I.start()
        # ~ I.wait()
        # ~ print("End third time")



if __name__ == "__main__":

    fname = "/scratch/paleo/bamboo/nabu.conf"

    T = OuterThread(fname)
    T.start()
    T.wait()





class Pipeline:
    def __init__(self, fname, subregion):
        self.proc = ProcessConfig(fname)
        # ~ self.proc.processing_steps.remove("save")
        self.pipeline = CudaFullFieldPipeline(self.proc, subregion)

    def process_chunk(self, sub_region=None):
        self.pipeline.ctx.push() #
        self.pipeline.process_chunk(sub_region=sub_region)
        # ~ self.pipeline.recs = self.pipeline._d_recs # if no "save"











from os import path
import pkgutil
import sys
import nabu

# module_path = path.dirname(mymodule.__file__)
def get_module_files(module_path, print_files=False):
    module_name = path.basename(module_path)
    res = {}
    for _, submodule_name, is_folder in pkgutil.iter_modules([module_path]):
        submodule_path = path.join(module_path, submodule_name)
        if is_folder:
            res[submodule_name] = get_module_files(submodule_path, print_files=print_files)
        fname = submodule_path + ".py"
        if path.isfile(fname):
            res[submodule_name] = fname
            if print_files:
                print(fname)
    return res



def get_module_output_name(module_path, fname, target_dir):
    fname2 = fname.replace(module_path, "")
    module_name = path.basename(module_path)
    return path.join(target_dir, module_name + fname2.replace("/", ".").replace(".py", ""))


if __name__ == "__main__":
    def print_usage():
        print("Usage: %s <get_files|get_path|get_module_output_name>" % sys.argv[0])

    nargs = len(sys.argv[1:])
    if nargs == 0:
        print_usage()
        exit(0)
    nabu_path = path.dirname(nabu.__file__)
    if sys.argv[1] == "get_files":
        get_module_files(nabu_path, print_files=True)
    elif sys.argv[1] == "get_path":
        print(nabu_path)
    elif sys.argv[1] == "get_module_output_name":
        if nargs != 3:
            print("Usage: %s get_module_outfile <file name> <target dir>" % sys.argv[0])
            exit(0)
        print(get_module_output_name(nabu_path, sys.argv[2], sys.argv[3]))
        exit(0)
    else:
        print_usage()
        exit(0)



def interp_sinos_mp(sinos, angles, nproc=1):
    n_a = angles.size
    angles_1 = angles[:n_a//2]
    angles_2 = angles[n_a//2:]
    sinos_interpolated = np.zeros(
        (sinos.shape[0], angles_2.size, sinos.shape[-1]), dtype=sinos.dtype
    )
    def _interp_sino(args):
        sino, sino_interpolated = args
        interpolator = interp1d(angles_2, sino[angles.size//2:, :], axis=0, fill_value="extrapolate")
        sino_interpolated[:] = interpolator(angles_1 + 180)
        return None
    with ThreadPool(nproc) as tp:
        tp.map(_interp_sino, zip(sinos, sinos_interpolated))
    return sinos_interpolated




def interp_sinos(sinos, angles):
    n_a = angles.size
    angles_1 = angles[:n_a//2]
    angles_2 = angles[n_a//2:]
    sinos_interpolated = np.zeros(
        (sinos.shape[0], angles_2.size, sinos.shape[-1]), dtype=sinos.dtype
    )
    for i in range(sinos.shape[0]):
        sino = sinos[i, n_a//2:, :]
        interpolator = interp1d(angles_2, sino, axis=0, fill_value="extrapolate")
        sinos_interpolated[i] = interpolator(angles_1 + 180)
    return sinos_interpolated



%%cython --compile-args=-fopenmp --link-args=-fopenmp -a
#%%cython -a
#cython: embedsignature=True, language_level=3, binding=True
#cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
## This is for developping:
## cython: profile=True, warn.undeclared=True, warn.unused=True, warn.unused_result=False, warn.unused_arg=True

import numpy as np
from cython.parallel import prange
"""
def bin2_cython(float[:, ::1] img):
    cdef:
        float[:, ::1] res
        int i, j, Ny, Nx
          
    shp = img.shape
    Ny, Nx = (img.shape[0]//2, img.shape[1]//2)
    res = np.zeros((Ny, Nx), dtype="f")
    for i in prange(Ny, nogil=True): 
        for j in range(Nx):
            res[i, j] = 0.25 * (img[2*i, 2*j] + img[2*i+1, 2*j] + img[2*i, 2*j+1] + img[2*i+1, 2*j+1])
    return np.asarray(res)
"""


def bin2_cython(unsigned short[:, ::1] img):
    cdef:
        float[:, ::1] res
        int i, j, Ny, Nx
          
    shp = img.shape
    Ny, Nx = (img.shape[0]//2, img.shape[1]//2)
    res = np.zeros((Ny, Nx), dtype="f")
    for i in prange(Ny, nogil=True): 
        for j in range(Nx):
            res[i, j] = 0.25 * (img[2*i, 2*j] + img[2*i+1, 2*j] + img[2*i, 2*j+1] + img[2*i+1, 2*j+1])
    return np.asarray(res)


#!/usr/bin/env python

from nabu.resources.nabu_config import nabu_config


def generate(file_):
    def write(content):
        print(content, file=file_)
    for section, values in nabu_config.items():
        if section == "about":
            continue
        write("## %s\n" % section)
        for key, val in values.items():
            if val["type"] == "unsupported":
                continue
            write(val["help"] + "\n")
            write(
                "```ini\n%s = %s\n```"
                % (key, val["default"])
            )



if __name__ == "__main__":

    import sys, os
    print(os.path.abspath(__file__))
    exit(0)

    fname = "/tmp/test.md"
    with open(fname, "w") as f:
        generate(f)



import numpy
import pytest

from nabu.stitching.overlap import compute_image_minimum_divergence, compute_image_higher_signal, check_overlaps
from nabu.testutils import get_data


def test_compute_image_minimum_divergence():
    """make sure the compute_image_minimum_divergence function is processing"""
    raw_data_1 = get_data("brain_phantom.npz")["data"]
    raw_data_2 = numpy.random.rand(*raw_data_1.shape) * 255.0

    stitching = compute_image_minimum_divergence(raw_data_1, raw_data_2, high_frequency_threshold=2)
    assert stitching.shape == raw_data_1.shape


def test_compute_image_higher_signal():
    """
    make sure compute_image_higher_signal is processing
    """
    raw_data = get_data("brain_phantom.npz")["data"]
    raw_data_1 = raw_data.copy()
    raw_data_1[40:75] = 0.0
    raw_data_1[:, 210:245] = 0.0

    raw_data_2 = raw_data.copy()
    raw_data_2[:, 100:120] = 0.0

    stitching = compute_image_higher_signal(raw_data_1, raw_data_2)

    numpy.testing.assert_array_equal(
        stitching,
        raw_data,
    )


def test_check_overlaps():
    """test 'check_overlaps' function"""

    # two frames, ordered and with an overlap
    check_overlaps(
        frames=(
            numpy.ones(10),
            numpy.ones(20),
        ),
        positions=((10,), (0,)),
        axis=0,
        raise_error=True,
    )

    # two frames, ordered and without an overlap
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(10),
                numpy.ones(20),
            ),
            positions=((0,), (100,)),
            axis=0,
            raise_error=True,
        )

    # two frames, frame 0 fully overlap frame 1
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(20),
                numpy.ones(10),
            ),
            positions=((8,), (5,)),
            axis=0,
            raise_error=True,
        )

    # three frames 'overlaping' as expected
    check_overlaps(
        frames=(
            numpy.ones(10),
            numpy.ones(20),
            numpy.ones(10),
        ),
        positions=((20,), (10,), (0,)),
        axis=0,
        raise_error=True,
    )

    # three frames: frame 0 overlap frame 1 but also frame 2
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(20),
                numpy.ones(10),
                numpy.ones(10),
            ),
            positions=((20,), (15,), (11,)),
            axis=0,
            raise_error=True,
        )



import numpy as np
import astra
from utils import ims


def _create_projector_parallel(vol_geom, angles, dx=1., dy=1.):
    n_x, n_z = vol_geom['GridColCount'], vol_geom['GridSliceCount']
    # det_spacing_x, det_spacing_y, det_row_count, det_col_count, angles
    proj_geom = astra.create_proj_geom('parallel3d', dx, dy, n_z, n_x, angles)
    projector = astra.create_projector('cuda3d', proj_geom, vol_geom)
    return projector, proj_geom


def _create_projector_tilted(vol_geom, n_angles, alpha):
    n_x, n_z = vol_geom['GridColCount'], vol_geom['GridSliceCount']
    vectors = np.zeros((n_angles, 12))
    angles = np.linspace(0, np.pi, n_angles)
    c_a = np.cos(alpha)
    s_a = np.sin(alpha)
    for i in range(n_angles):
        # ray direction
        vectors[i,0] = np.sin(angles[i]) * c_a
        vectors[i,1] = -np.cos(angles[i]) * c_a
        vectors[i,2] = s_a

        # center of detector
        # ~ vectors[i,3] = 0
        # ~ vectors[i,4] = 0
        # ~ vectors[i,5] = 0

        # vector from detector pixel (0,0) to (0,1)
        vectors[i,6] = np.cos(angles[i])
        vectors[i,7] = np.sin(angles[i])
        vectors[i,8] = 0

        # vector from detector pixel (0,0) to (1,0)
        vectors[i,9] = -np.sin(angles[i]) * s_a
        vectors[i,10] = np.cos(angles[i]) * s_a
        vectors[i,11] = c_a

    proj_geom = astra.create_proj_geom('parallel3d_vec', n_z, n_x, vectors)
    projector = astra.create_projector('cuda3d', proj_geom, vol_geom)
    return projector, proj_geom


def proj3D(volume, n_a):
    n_x, n_y, n_z = volume.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z) # rows, colums, slices (y, x, z)
    vol_astra = astra.data3d.create('-vol', vol_geom, volume)

    angles = np.linspace(0, np.pi, n_a)
    projector, proj_geom = _create_projector_parallel(vol_geom, angles)

    projs_id, projs = astra.create_sino3d_gpu(vol_astra, proj_geom, vol_geom)
    astra.data3d.clear() #

    return projs


def proj3D_tilt(volume, n_a, alpha):
    n_x, n_y, n_z = volume.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z) # rows, colums, slices (y, x, z)
    vol_astra = astra.data3d.create('-vol', vol_geom, volume)

    projector, proj_geom = _create_projector_tilted(vol_geom, n_a, alpha)

    projs_id, projs = astra.create_sino3d_gpu(vol_astra, proj_geom, vol_geom)
    astra.data3d.clear() #

    return projs





def create_projectors(vol, n_angles, alpha):
    n_x, n_y, n_z = vol.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z) # rows, colums, slices (y, x, z)

    vectors = np.zeros((n_angles, 12))
    angles = np.linspace(0, np.pi, n_angles)
    c_a = np.cos(alpha)
    s_a = np.sin(alpha)
    for i in range(n_angles):
        # ray direction
        vectors[i,0] = np.sin(angles[i]) * c_a
        vectors[i,1] = -np.cos(angles[i]) * c_a
        vectors[i,2] = s_a

        # center of detector
        # ~ vectors[i,3] = 0
        # ~ vectors[i,4] = 0
        # ~ vectors[i,5] = 0

        # vector from detector pixel (0,0) to (0,1)
        vectors[i,6] = np.cos(angles[i])
        vectors[i,7] = np.sin(angles[i])
        vectors[i,8] = 0

        # vector from detector pixel (0,0) to (1,0)
        vectors[i,9] = -np.sin(angles[i]) * s_a
        vectors[i,10] = np.cos(angles[i]) * s_a
        vectors[i,11] = c_a

    proj_geom = astra.create_proj_geom('parallel3d_vec', n_z, n_x, vectors)

    vol_id = astra.data3d.create('-vol', vol_geom, data=vol)
    # ~ proj_id = astra.data3d.create('-sino', proj_geom)

    # ~ proj_id = astra.create_projector('cuda3d', proj_geom, vol_geom)
    # ~ sinogram_id, sinogram = astra.create_sino3d_gpu(vol, proj_id)
    sino_id, sino = astra.create_sino3d_gpu(vol, proj_geom, vol_geom, returnData=True)

    # ~ rec_id, rec = astra.create_backprojection3d_gpu(sino, proj_geom, vol_geom, returnData=True)

    rec_id = astra.data3d.create('-vol', vol_geom)

    # Set up the parameters for a reconstruction algorithm using the GPU
    # ~ cfg = astra.astra_dict('SIRT3D_CUDA')
    cfg = astra.astra_dict('BP3D_CUDA')
    cfg['ReconstructionDataId'] = rec_id
    cfg['ProjectionDataId'] = sino_id
    alg_id = astra.algorithm.create(cfg)

    astra.algorithm.run(alg_id, 10)

    rec = astra.data3d.get(rec_id)






def fbp3D_tilted(projector, sino):
    cfg_backproj = astra.creators.astra_dict('FBP3D_CUDA')
    cfg_backproj['FilterType'] = 'Ram-Lak'
    cfg_backproj['ProjectorId'] = projector
    pg = astra.projector.projection_geometry(projector)
    # In
    sid = astra.data3d.link('-sino', pg, sino)
    cfg_backproj['ProjectionDataId'] = sid
    # Out
    vg = astra.projector.volume_geometry(projector)
    vshape = astra.functions.geom_size(vg)
    v = np.zeros(vshape, dtype=np.float32)
    vid = astra.data2d.link('-vol', vg, v)
    cfg_backproj['ReconstructionDataId'] = vid

    bp_id = astra.algorithm.create(cfg_backproj)
    astra.algorithm.run(bp_id)
    astra.algorithm.delete(bp_id)
    astra.data3d.delete([sid, vid])
    return v







def example():
    vol = np.load("/scisoft/users/paleo/data/maquette/MRI512cube.npy")

    tilt = np.deg2rad(20)
    n_angles = 180
    projs = proj3D_tilt(vol, n_angles, tilt)
    projs0 = proj3D(vol, n_angles)


    legend=["Proj 0", "Proj 45", "Proj 90", "Proj 135"]
    print("Standard parallel geometry")
    ims(
        [projs0[:, 0, :], projs0[:, 45, :], projs0[:, 90, :], projs0[:, 135, :]],
        legend=legend
    )
    print("Parallel geometry with tilted rotation axis")
    ims(
        [projs[:, 0, :], projs[:, 45, :], projs[:, 90, :], projs[:, 135, :]],
        legend=legend
    )









def mc_tilt_vals(shape, n_a, alpha, n_sim=5):
    n_z, n_y, n_x = shape
    x0 = (n_x - 1)/2.
    y0 = (n_y - 1)/2.
    angles = np.linspace(0, np.pi, n_a, False)
    s_a = np.sin(alpha)
    c_a = np.cos(alpha)
    s_t = np.sin(angles)
    c_t = np.cos(angles)
    vals = {}
    for i in range(n_a):
        for j in range(n_sim):
            x = np.random.randint(0, high=n_x)
            y = np.random.randint(0, high=n_y)
            z = np.random.randint(0, high=n_z)
            vals[(x, y, z)] = s_a * (s_t[i] * (x-x0) + x0) \
            + s_a * (c_t[i] * (y-y0) + y0) \
            + c_a * z
    return vals

"""
#
# For 512**3 with angular range [0, 2pi] and alpha = 50 mrad,
# we have delta_z_max = 40
#
V = mc_tilt_vals((512, 512, 512), 180, 50e-3, n_sim=100)
Z = np.array(list(map(lambda x: x[-1], V.keys())))
hist(np.array(list(V.values())) - Z, bins=64)

# For 14 mrad, delta_z_max = 13
# For 14 mrad and volume 2048**3, we have delta_z_max = 50

"""



"""
def get_sino_tilt(sinos, alpha, rot_center=None):
    n_z, n_a, n_x = sinos.shape
    if rot_center is None:
        rot_center = (n_x - 1)/2.
    c_a = np.cos(alpha)
    s_a = np.sin(alpha)

    X = np.arange(n_x)
    Y = np.arange(n_y)
    X *
"""

















import numpy as np
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.pool import ThreadPool
from multiprocessing import Pool
from itertools import product
from functools import partial

# Constants
# Miller index of reflection
H = 0
# Miller index of reflection
K = 4
# Number of unit cells per direction
N = 32
# Defines how many points are needed to describe a single Laue fringe (2 = Nyquist frequency)
oversampling = 3

# Radius of the crystal
R = N/2

# Maximum strain at surface
e0 = 0.01
# Width of the strain profile below the surface
w = 5.

# Generate real and reciprocal space coordinates
n = np.arange(N)
m = np.arange(N)
h = np.arange(H-0.5, H+0.5, 1./(oversampling*N))
k = np.arange(K-0.5, K+0.5, 1./(oversampling*N))


# Displacement of atoms as function of the radius
def delta(radius, crystal_radius, strain_width):
    """Displacement of atoms as function of the radius"""
    return 1 + np.tanh((radius - crystal_radius) / strain_width)



def circ_numpy_2(N, h, k):
    R, C = np.indices((N, N))
    radius = np.sqrt((R - N/2.)**2 + (C - N/2.)**2)
    support = radius**2 <= (N/2)**2
    Delta = delta(radius, N/2, w)
    delta_n = e0 * (n - N/2) * Delta
    delta_n_n = n + delta_n
    delta_m_m = delta_n_n.T

    res = np.zeros((h.size, k.size))
    for hk in product(enumerate(h), enumerate(k)):
        (i_h, v_h), (i_k, v_k) = hk
        res[i_h, i_k] = np.abs(np.sum(
            support * np.exp(2j * np.pi * (v_h*delta_n_n + v_k*delta_m_m))
        ))**2
    return res


def compute_image_part(my_part, N, h, k):
    (start_x, end_x), (start_y, end_y) = my_part

    # Retrieve shared arrays
    my_s_support = SharedMemory(create=False, name="support")
    my_support = np.ndarray((N, N), dtype=np.float64, buffer=my_s_support.buf)

    my_s_delta_n_n = SharedMemory(create=False, name="delta_n_n")
    my_delta_n_n = np.ndarray((N, N), dtype=np.float64, buffer=my_s_delta_n_n.buf)
    my_delta_m_m = my_delta_n_n.T

    my_s_result = SharedMemory(create=False, name="result")
    my_result = np.ndarray((h.size, k.size), dtype=np.float64, buffer=my_s_result.buf)

    # Compute on a partial set of reciprocal coordinates
    for i_h, i_k in product(range(start_y, end_y), range(start_x, end_x)):
        v_h = h[i_h]
        v_k = k[i_k]
        my_result[i_h, i_k] = np.abs(np.sum(my_support * np.exp(2j * np.pi * (v_h*my_delta_n_n + v_k*my_delta_m_m))))**2


def circ_mp(N, h, k):
    R, C = np.indices((N, N))
    radius = np.sqrt((R - N/2.)**2 + (C - N/2.)**2)
    # support = radius**2 <= (N/2)**2
    Delta = delta(radius, N/2, w)
    delta_n = e0 * (n - N/2) * Delta
    # delta_n_n = n + delta_n
    # delta_m_m = delta_n_n.T

    parts_1D = [(0, h.size//2), (h.size//2, h.size)]
    parts = list(product(parts_1D, parts_1D))

    try:
        # Create shared arrays
        s_support = SharedMemory(create=True, size=N*N*np.dtype(np.float64).itemsize, name="support")
        support = np.ndarray((N, N), dtype=np.float64, buffer=s_support.buf)
        support[:] = radius**2 <= (N/2)**2

        s_delta_n_n = SharedMemory(create=True, size=support.nbytes, name="delta_n_n")
        delta_n_n = np.ndarray((N, N), dtype=np.float64, buffer=s_delta_n_n.buf)
        delta_n_n[:] = n + delta_n

        s_result = SharedMemory(
            create=True, size=h.size * k.size * np.dtype(np.float64).itemsize, name="result"
        )
        result = np.ndarray((h.size, k.size), dtype=np.float64, buffer=s_result.buf)
        result[:] = 0.

        # Dispatch computations
        with Pool(4) as p:
            p.map(partial(compute_image_part, N=N, h=h, k=k), parts)

    finally:
        res = result.copy()
        for s_array in [s_support, s_delta_n_n, s_result]:
            s_array.unlink()

    return res







def circ_mt(N, h, k):
    R, C = np.indices((N, N))
    radius = np.sqrt((R - N/2.)**2 + (C - N/2.)**2)
    support = radius**2 <= (N/2)**2
    Delta = delta(radius, N/2, w)
    delta_n = e0 * (n - N/2) * Delta
    delta_n_n = n + delta_n
    delta_m_m = delta_n_n.T

    parts_1D = [(0, h.size//2), (h.size//2, h.size)]
    parts = list(product(parts_1D, parts_1D))

    result = np.zeros((h.size, k.size), dtype=np.float64)

    def compute_image_part(my_part):
        (start_x, end_x), (start_y, end_y) = my_part
        # Compute on a partial set of reciprocal coordinates
        for i_h, i_k in product(range(start_y, end_y), range(start_x, end_x)):
            v_h = h[i_h]
            v_k = k[i_k]
            result[i_h, i_k] = np.abs(np.sum(
                support * np.exp(2j * np.pi * (v_h*delta_n_n + v_k*delta_m_m))
            ))**2

    with ThreadPool(4) as tp:
        tp.map(compute_image_part, parts)


    return result
















from matplotlib.pyplot import subplots
from matplotlib.colors import LogNorm
def display(result):
    "Display the array"
    fig, ax = subplots()
    fig.suptitle("Bragg peak")
    ax.imshow(result.T, extent=(h.min(), h.max(), k.min(), k.max()), norm=LogNorm(), origin = 'lower')
    ax.set_xlabel('H');
    ax.set_ylabel('K')
    ax.set_title("Crystal")



def circ_einsum(N, h, k):
    R, C = np.indices((N, N))
    radius = np.sqrt((R - N/2.)**2 + (C - N/2.)**2)
    support = radius**2 <= (N/2)**2
    Delta = delta(radius, N/2, w)
    delta_n = e0 * (n - N/2) * Delta
    delta_n_n = n + delta_n
    delta_m_m = delta_n_n.T

    tmp1 = np.exp(2j*pi*np.einsum("i,jk", h, delta_n_n))
    tmp2 = np.exp(2j*pi*np.einsum("i,jk", k, delta_m_m))
    res = np.abs(np.einsum("jk,ijk,ljk->il", support, tmp1, tmp2))**2

    return res



from posixpath import join
import numpy as np
from nabu.io.utils import get_h5_value
from nabu.app.utils import parse_params_values
from nxtomomill.nexus.nxtomo import NXtomo
from nxtomomill.utils import ImageKey
from integrator.hdf5 import ID15Dataset, list_datasets

def convert_id15_dataset(
    fname, output_fname,
    detector_name="pcoedgehs", ignore_last_n_projections=100,
    energy=None, distance=None
):

    entries = list_datasets(fname)
    if len(entries) < 3:
        print("Error: Expected at least three datasets, got only %d" % (len(entries)))
        return
    do_360 = False
    if len(entries) == 6:
        do_360 = True

    projs = ID15Dataset(fname, detector_name=detector_name, entry="1.1")
    flats = ID15Dataset(fname, detector_name=detector_name, entry="2.1")
    darks = ID15Dataset(fname, detector_name=detector_name, entry="3.1")
    if do_360:
        projs2 = ID15Dataset(fname, detector_name=detector_name, entry="4.1")
        flats2 = ID15Dataset(fname, detector_name=detector_name, entry="5.1")
        darks2 = ID15Dataset(fname, detector_name=detector_name, entry="6.1")

    my_nxtomo = NXtomo("entry")

    data_urls = [
        darks.dataset_hdf5_url,
        flats.dataset_hdf5_url,
        projs.dataset_hdf5_url
    ]
    if do_360:
        data_urls.extend([
            # darks2.dataset_hdf5_url,
            flats2.dataset_hdf5_url,
            projs2.dataset_hdf5_url
        ])
    my_nxtomo.instrument.detector.data = data_urls

    img_keys = [
        [ImageKey.DARK_FIELD] * darks.data_shape[0],
        [ImageKey.FLAT_FIELD] * flats.data_shape[0],
        [ImageKey.PROJECTION] * (projs.data_shape[0] - ignore_last_n_projections),
        [ImageKey.ALIGNMENT]  * ignore_last_n_projections,
    ]
    if do_360:
        img_keys.extend([
            # [ImageKey.DARK_FIELD] * darks2.data_shape[0],
            [ImageKey.FLAT_FIELD] * flats2.data_shape[0],
            [ImageKey.PROJECTION] * (projs2.data_shape[0] - ignore_last_n_projections),
            [ImageKey.ALIGNMENT]  * ignore_last_n_projections,
        ])

    my_nxtomo.instrument.detector.image_key_control = np.concatenate(img_keys)

    rotation_angles = get_h5_value(projs.fname, join(projs.entry, "instrument/hrrz/data"))
    if rotation_angles is not None:
        last_idx = None
        if ignore_last_n_projections > 0:
            last_idx = -ignore_last_n_projections
        rotation_angles = [
            [0] * darks.data_shape[0],
            [0] * flats.data_shape[0],
            rotation_angles[:last_idx],
            [0] * ignore_last_n_projections
        ]
        if do_360:
            rotation_angles2 = get_h5_value(projs2.fname, join(projs2.entry, "instrument/hrrz/data"))
            rotation_angles.extend([
                # [0] * darks2.data_shape[0],
                [0] * flats2.data_shape[0],
                rotation_angles2[:last_idx],
                [0] * ignore_last_n_projections
            ])
        print(np.concatenate(rotation_angles).shape)
        my_nxtomo.sample.rotation_angle = np.concatenate(rotation_angles)

    # my_nxtomo.instrument.detector.field_of_view = "Full" if not(do_360) else "Half"
    my_nxtomo.instrument.detector.field_of_view = "Full"
    my_nxtomo.instrument.detector.x_pixel_size = my_nxtomo.instrument.detector.y_pixel_size = 6.5 * 1e-6

    if energy is not None:
        my_nxtomo.energy = energy  # in keV by default
    if distance is not None:
        my_nxtomo.instrument.detector.distance = distance  # in meter


    my_nxtomo.save(file_path=output_fname, overwrite=True, nexus_path_version=1.1)





CLIConfig = {
    "input_file": {
        "help": "Bliss HDF5 file",
        "default": "",
        "mandatory": True,
    },
    "output_file": {
        "help": "Output NX file",
        "default": "",
        "mandatory": True,
    },
    "ignore_last_n_projections": {
        "help": "Number of projections to ignore in the end of each series of projections.",
        "default": 0,
        "type": int,
    },
    "detector_name": {
        "help": "Detector name. Default is pcoedgehs",
        "default": "pcoedgehs",
    },
    "energy": {
        "help": "Incident beam energy in keV",
        "default": 0.,
        "type": float,
    },
    "distance": {
        "help": "Sample-detector distance in meters",
        "default": 0.,
        "type": float,
    },
}


if __name__ == "__main__":

    args = parse_params_values(
        CLIConfig,
        parser_description="Convert a ID15A absorption tomography acquisition file to NX format",
    )

    convert_id15_dataset(
        args["input_file"],
        args["output_file"],
        detector_name=args["detector_name"],
        ignore_last_n_projections=args["ignore_last_n_projections"],
        energy=args["energy"] or None,
        distance=args["distance"] or None
    )




import numpy as np
from math import sqrt, pi

from nabu.utils import updiv, get_cuda_srcfile, _sizeof, nextpow2, convert_index
from nabu.cuda.utils import copy_array
from nabu.cuda.kernel import CudaKernel
import pycuda.driver as cuda
from pycuda import gpuarray as garray


# PATCH
import os
def get_cuda_srcfile(filename):
    # ~ cuda_src_folder = "/home/pierre/workspace/git/nabu/sandbox"
    cuda_src_folder = "/mntdirect/_scisoft/users/paleo/tmp/nabu_test/sandbox"
    return os.path.join(cuda_src_folder, filename)


class SinoTilter:
    def __init__(
        self,
        sino_shape,
        axis_tilt,
        angles=None,
        rot_center=None,
        axis_tilt_center=None
    ):

        self._init_geometry(sino_shape, angles, rot_center, axis_tilt, axis_tilt_center)
        self._allocate_memory()
        self._compute_angles()
        self._compile_kernels()
        self._bind_textures()


    def _init_geometry(self, sino_shape, angles, rot_center, axis_tilt, axis_tilt_center):
        self.sino_shape = sino_shape
        if len(sino_shape) == 3:
            n_slices, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Expected 3D sinogram")
        n_sinos = n_slices
        self.dwidth = dwidth
        self.n_slices = n_slices
        self._set_slice_shape(None)
        self.n_sinos = n_sinos
        self.rot_center = rot_center or (self.dwidth - 1)/2.
        self.axis_pos = self.rot_center
        self._set_angles(angles, n_angles)
        if axis_tilt is None:
            axis_tilt = 0
        self.axis_tilt = axis_tilt
        if axis_tilt_center is None:
            y1 = (self.n_y - 1)/2.
            z1 = self.n_slices
            axis_tilt_center = (y1, z1)
        self.axis_tilt_center = axis_tilt_center


    def _set_slice_shape(self, slice_shape):
        n_y = self.dwidth
        n_x = self.dwidth
        if slice_shape is not None:
            if np.isscalar(slice_shape):
                slice_shape = (slice_shape, slice_shape)
            n_y, n_x = slice_shape
        self.n_x = n_x
        self.n_y = n_y
        self.slice_shape = (n_y, n_x)
        if self.n_slices > 1:
            self.slice_shape = (self.n_slices,) + self.slice_shape


    def _set_angles(self, angles, n_angles):
        self.n_angles = n_angles
        if angles is None:
            angles = n_angles
        if np.isscalar(angles):
            angles = np.linspace(0, np.pi, angles, False)
        else:
            assert len(angles) == self.n_angles
        self.angles = angles


    def _allocate_memory(self):
        self._d_sino_cua = cuda.np_to_array(np.zeros(self.sino_shape, "f"), "C")
        # 1D textures are not supported in pycuda
        self.h_sin = np.zeros((1, self.n_angles), "f")
        self.h_cos = np.zeros((1, self.n_angles), "f")
        self._d_sino = garray.zeros(self.sino_shape, "f")
        self._d_out = garray.zeros(self.sino_shape[1:], "f")
        self._d_proc = garray.zeros(self.sino_shape[1:], np.int32)


    def _compute_angles(self):
        self.h_cos[0] = np.cos(self.angles).astype("f")
        self.h_sin[0] = np.sin(self.angles).astype("f")
        self._d_sin = garray.to_gpu(self.h_sin[0])
        self._d_cos = garray.to_gpu(self.h_cos[0])


    def _get_kernel_options(self):
        sourcemodule_options = []
        block = (16, 16, 1)
        grid = (
            updiv(self.n_x, block[0]),
            updiv(self.n_y, block[1]),
            1
        )
        sourcemodule_options.append("-DBACKPROJ3D")
        shared_size = int(np.prod(block)) * 2
        shared_size *= 4 # sizeof(float32)
        self._kernel_options = {
            "file_name": get_cuda_srcfile("sino_tilt.cu"),
            "kernel_name": "sino_tilt",
            "kernel_signature": "PiiiifPPffffP",
            "texture_name": "tex_projections3D",
            "sourcemodule_options": sourcemodule_options,
            "grid": grid,
            "block": block,
            "shared_size": shared_size
        }


    def _compile_kernels(self):
        self._get_kernel_options()
        kern_opts = self._kernel_options
        # Configure backprojector
        self.gpu_projector = CudaKernel(
            kern_opts["kernel_name"],
            filename=kern_opts["file_name"],
            options=kern_opts["sourcemodule_options"]
        )
        self.texref_proj = self.gpu_projector.module.get_texref(
            kern_opts["texture_name"]
        )
        self.texref_proj.set_filter_mode(cuda.filter_mode.LINEAR)
        self.gpu_projector.prepare(kern_opts["kernel_signature"], [self.texref_proj])
        # Prepare kernel arguments
        self.kern_proj_args = [
            self._d_out,
            None, # z
            self.n_angles,
            self.dwidth,
            self.n_slices,
            self.axis_pos,
            self._d_cos,
            self._d_sin,
            np.cos(self.axis_tilt),
            np.sin(self.axis_tilt),
            self.axis_tilt_center[0],
            self.axis_tilt_center[1],
            self._d_proc
        ]
        self.kern_proj_kwargs = {
            "grid": kern_opts["grid"],
            "block": kern_opts["block"],
            "shared_size": kern_opts["shared_size"],
        }


    def _bind_textures(self):
        self.texref_proj.set_array(self._d_sino_cua)


    def transform(self, sino, z, output=None, do_checks=True):
        copy_array(self._d_sino_cua, sino, check=do_checks)
        self.kern_proj_args[1] = z
        if output is not None:
            self.kern_proj_args[0] = output # output has to be a gpuarray
        self.gpu_projector(
            *self.kern_proj_args,
            **self.kern_proj_kwargs
        )
        if output is not None:
            self.kern_proj_args[0] = self._d_out
            return output
        else:
            return self._d_out.get()





from nabu.preproc.ctf_id16 import ID16CTFPhaseRetrieval
from nabu.preproc.ctf_id16 import ID16CTFPhaseRetrieval
from nabu.preproc.ccd import FlatFieldArrays
from nabu.preproc.ctf import GeoPars
from nabu.testutils import get_data
test_data = get_data("ctf_tests_data_all_pars.npz")
rand_disp_vh = test_data["rh"]
rand_disp_vh.shape = [rand_disp_vh.shape[0], rand_disp_vh.shape[2]]
geo_pars = GeoPars(
	z1_vh=test_data["z1_vh"],
	z2=test_data["z2"],
	pix_size_det=test_data["pix_size_det"],
	length_scale=test_data["length_scale"],
	wavelength=test_data["wave_length"],
)
flats = [test_data["ref0"], test_data["ref1"]]
dark = test_data["dark"]

flatfield = FlatFieldArrays(
    [1200] + list(test_data["img_shape_vh"]), {0: flats[0], 1200: flats[1]}, {0: dark}
)
ctf_filter = ID16CTFPhaseRetrieval(
	geo_pars, 27.0,
	lim1=1.0e-5, lim2=0.2,
	flatfielder=flatfield,
	flat_distorsion_params={
		"tile_size": 100,
		"interpolation_kind": "cubic",
		"padding_mode": "edge",
		"correction_spike_threshold": 3.0
	},
	spikes_removal_threshold=0.04
)
img = test_data["im"].astype("f")
phase = ctf_filter.retrieve_phase(
	img, int(test_data["ipro"]),
	padded_shape=test_data["padded_img_shape_vh"],
	translation_vh=rand_disp_vh[:, int(test_data["ipro"])],
	normalize_by_mean=True
)



import numpy as np

def linear_interp(arr, x, new_x):
    dx = np.diff(x)
    dy = np.diff(arr)
    if new_x[0] > x[0]:
        dx = np.hstack([dx, dx[-1]])
        dy = np.hstack([dy, dy[-1]])
    else:
        dx = np.hstack([dx[0], dx])
        dy = np.hstack([dy[0], dy])        
    return (dy / dx) * (new_x - x) + arr
    
  



from nabu.resources.computations import estimate_chunk_size
from nabu.preproc.phase import compute_paganin_margin
from nabu.cuda.utils import get_gpu_memory
from nabu.resources.processconfig import ProcessConfig
from nabu.app.fullfield_cuda import CudaFullFieldPipeline, CudaFullFieldPipelineLimitedMemory


"""
fname = "/scratch/paleo/bamboo/nabu.conf"
z_start = 0
z_end = 500
gpu_id = 0
gpu_mem_fraction = 0.9



proc = ProcessConfig(fname)
# For bamboo
proc.dataset_infos.dataset_scanner._energy = 19.
proc.processing_options["phase"]["energy_kev"] = 19.



chunk_size = estimate_chunk_size(
    get_gpu_memory(gpu_id) * gpu_mem_fraction,
    process_config,
    chunk_step=10
)

if "phase" in proc.processing_steps:
    margin_v, margin_h = compute_paganin_margin(
        radio_shape,
        distance=opts["distance_cm"],
        energy=opts["energy_kev"],
        delta_beta=opts["delta_beta"],
        pixel_size=opts["pixel_size_microns"],
        padding=opts["padding_type"]
    )



F = CudaFullFieldPipeline(proc, (start_z, start_z + delta_z))
#F = CudaFullFieldPipelineLimitedMemory(proc, (start_z, start_z + delta_z), chunk_size=chunk_size

"""

"""

Logic/flow
------------

get the requested GPU, or take "0"

estimate chunk size
estimate paganin margin

if paganin margin > chunk size:
    CudaFullFieldPipelineLimitedMemory
default: CudaFullFieldPipeline

if (z_end - z_start) > chunk_size:
    cut by chunks, with overlap
else:
    chunk_size = (z_end - z_start) + margin

spawn pipeline
exec


Architecture
-------------

WorkersManager:
  Gère la logique: ressources, découpage en chunks, ...
  Instancie les Workers (local, slurm, ...)


"""

from math import ceil
from psutil import virtual_memory
from nabu.resources.logger import LoggerOrPrint
from nabu.cuda.utils import collect_cuda_gpus


def get_gpus_ids(resources_cfg):
    gpus_ids = resources_cfg["gpu_id"]
    if gpus_ids != []:
        return gpus_ids
    # TODO (?) heuristic to pick the best gpus
    return list(range(resources_cfg["gpus"]))



class SimpleWorkersManager:

    gpu_mem_fraction = 0.9
    cpu_mem_fraction = 0.9
    phase_processing_mode = "accurate"
    pipeline_cls = CudaFullFieldPipeline


    def __init__(self, process_config, logger=None):
        self.process_config = process_config
        self.logger = LoggerOrPrint(logger)
        self._get_reconstruction_range()
        self._get_resources()
        self._compute_chunk_size()
        self._compute_phase_margin()
        self._compute_volume_chunks()


    def _get_reconstruction_range(self):
        rec_cfg = self.process_config.nabu_config["reconstruction"]
        self.z_min = rec_cfg["start_z"]
        self.z_max = rec_cfg["end_z"]
        self.delta_z = self.z_max - self.z_min


    def _get_resources(self):
        self.resources = {}
        self._get_gpu()
        self._get_memory()


    def _get_memory(self):
        vm = virtual_memory()
        self.resources["mem_avail_GB"] = vm.available / 1e9


    def _get_gpu(self):
        gpus = get_gpus_ids(self.process_config.nabu_config["resources"])
        if len(gpus) == 0:
            raise ValueError("Need at least one GPU")
        if len(gpus) > 1:
            raise ValueError("This class does not support more than one GPU")
        self.resources["gpu_id"] = self._gpu_id = gpus[0]
        self.resources["gpus"] = collect_cuda_gpus()


    def _compute_chunk_size(self):
        gpu_mem = self.resources["gpus"][self._gpu_id]["memory_GB"] * self.gpu_mem_fraction
        cpu_mem = self.resources["mem_avail_GB"] * self.cpu_mem_fraction
        mem_avail = min(gpu_mem, cpu_mem)
        self.max_chunk_size = estimate_chunk_size(
            mem_avail,
            self.process_config,
            chunk_step=10
        )


    def _compute_phase_margin(self):
        if "phase" not in proc.processing_steps:
            self._phase_margin = (0, 0)
            return
        radio_shape = self.process_config.dataset_infos.radio_dims[::-1]
        opts = self.process_config.processing_options["phase"]
        margin_v, margin_h = compute_paganin_margin(
            radio_shape,
            distance=opts["distance_cm"],
            energy=opts["energy_kev"],
            delta_beta=opts["delta_beta"],
            pixel_size=opts["pixel_size_microns"],
            padding=opts["padding_type"]
        )
        self._phase_margin = (margin_v, margin_h)


    def _compute_volume_chunks(self):
        n_z = self.process_config.dataset_infos._radio_dims_notbinned[1]
        margin_v = self._phase_margin[0]
        self._margin_far_up = min(margin_v, self.z_min)
        self._margin_far_down = min(margin_v, n_z - (self.z_max + 1))

        # Actually less in some cases (margin_far_up + margin_far_down instead of 2*margin_v).
        # But we want to use only one class for all stages.
        chunk_size_for_one_slice = 1 + 2 * margin_v
        if chunk_size_for_one_slice > self.max_chunk_size:
            self.logger.warning(
                "Phase margin is %d, so we need to process at least %d detector rows. However, the available memory enables to process only %d rows at once"
                % (margin_v, chunk_size_for_one_slice, self.max_chunk_size)
            )
            if self.phase_processing_mode == "accurate": # TODO normalize the names
                self._pipeline_cls = CudaFullFieldPipelineLimitedMemory
                self.logger.debug("Using CudaFullFieldPipelineLimitedMemory")
            else:
                self._phase_margin = (0, 0)
                self._pipeline_cls = CudaFullFieldPipeline
                self.logger.debug("Using CudaFullFieldPipeline without margin")
        # | margin_up |     n_slices    |  margin_down |
        # |-----------|-----------------|--------------|
        # |----------------------------------------------------|
        #                    delta_z
        n_slices = self.max_chunk_size - (2 * margin_v)
        tasks = []
        if n_slices >= self.delta_z:
            n_slices = self.delta_z
            (margin_up, margin_down) = (self._margin_far_up, self._margin_far_down)
            tasks.append({
                "sub_region": (self.z_min - margin_up, self.z_max + margin_down),
                "phase_margin": ((margin_up, margin_down), (0, 0))
            })
        else:
            # In this case there are at least two stages
            n_stages = ceil(self.delta_z / n_slices)
            tasks = []
            curr_z_min = self.z_min
            curr_z_max = self.z_min + n_slices
            # First stage
            (margin_up, margin_down) = (self._margin_far_up, margin_v)
            tasks.append({
                "sub_region": (curr_z_min - margin_up, curr_z_max + margin_down),
                "phase_margin": ((margin_up, margin_down), (0, 0))
            })
            curr_z_min += n_slices
            curr_z_max += n_slices

            # Middle stages
            for i in range(1, n_stages - 1):
                (margin_up, margin_down) = (margin_v, margin_v)
                tasks.append({
                    "sub_region": (curr_z_min - margin_up, curr_z_max + margin_down),
                    "phase_margin": ((margin_up, margin_down), (0, 0))
                })
                curr_z_min += n_slices
                curr_z_max += n_slices
            # Last stage
            curr_z_max = min(curr_z_max, self.z_max)
            (margin_up, margin_down) = (margin_v, self._margin_far_down)
            tasks.append({
                "sub_region": (curr_z_min - margin_up, curr_z_max + margin_down),
                "phase_margin": ((margin_up, margin_down), (0, 0))
            })

        self.tasks = tasks
        self.n_slices = n_slices













import numpy as np
from nabu.estimation.translation import DetectorTranslationAlongBeam
from tomoscan.io import HDF5File


def get_data(fname, h5_paths):
    with HDF5File(fname, "r") as f:
        align_data = f[h5_paths["align_data"]][()]
        dark = f[h5_paths["dark"]][0]
        pixel_size_um = f[h5_paths["pixel_size"]][()]
        positions_mm = f[h5_paths["positions"]][()]

    # Dark subtraction
    align_data = align_data - dark.astype("f")

    return align_data, pixel_size_um, positions_mm



def get_alignment(align_data, pixel_size_um, positions_mm, plot=True):

    tr_calc = DetectorTranslationAlongBeam()
    if plot == True:
        tr_calc.verbose=True

    shifts_v, shifts_h = tr_calc.find_shift(align_data, positions_mm)

    # pixel_size is in microns, motor position is in mm
    tilt_v_deg = np.rad2deg(np.arctan(shifts_v * pixel_size_um / 1e3))
    tilt_h_deg = np.rad2deg(np.arctan(shifts_h * pixel_size_um / 1e3))
    print (f"\nVertical tilt  to be applied in deg (thy): {tilt_v_deg}")
    print (f"Horizontal tilt to be applied in deg (thz): {tilt_h_deg}\n")

    return align_data, shifts_v, shifts_h, positions_mm






if __name__ == "__main__":
    fname = "/data/id19/inhouse/id192201/id19/sample/sample_0001/sample_0001.h5"
    h5_paths = {
        "align_data": "11.1/measurement/pcolinux",
        "dark": "12.1/measurement/pcolinux",
        "pixel_size": "11.1/instrument/pcolinux/x_pixel_size",
        "positions": "11.1/measurement/hrxc"
    }
    align_data, pixel_size_um, positions_mm = get_data(fname, h5_paths)
    # align_data, shifts_v, shifts_h, positions_mm = get_alignment(align_data, pixel_size_um, positions_mm)

    D = DetectorTranslationAlongBeam()
    D. verbose = True
    sv, sh = D.find_shift(align_data, positions_mm)
    D.verbose = False
    sv, sh = D.find_shift(align_data, positions_mm)
    D.verbose = True
    sv, sh = D.find_shift(align_data, positions_mm)






import numpy as np
from math import sqrt, pi

from ..utils import updiv, get_cuda_srcfile, _sizeof, nextpow2, convert_index
from ..cuda.utils import copy_array
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from .filtering import SinoFilter
import pycuda.driver as cuda
from pycuda import gpuarray as garray
from .fbp import Backprojector

class PolarBackprojector(Backprojector):
    """
    Cuda Backprojector with output in polar coordinates.
    """

    cuda_fname = "backproj.cu"
    cuda_kernel_name = "backproj_polar"


    # patch parent method: force slice_shape to (n_angles, n_x)
    def _set_angles(self, angles, n_angles):
        Backprojector._set_angles(self, angles, n_angles)
        self.slice_shape = (self.n_angles, self.n_x)

    # patch parent method:
    def _set_slice_roi(self, slice_roi):
        if slice_roi is not None:
            raise ValueError("slice_roi is not supported with this class")
        Backprojector._set_slice_roi(self, slice_roi)

    # patch parent method: update kernel args
    def _compile_kernels(self):
        n_y = self.n_y
        self.n_y = self.n_angles
        Backprojector._compile_kernels()
        self.n_y = n_y




from time import sleep
import pycuda.autoinit
import pycuda.gpuarray as garray
from nabu.resources.processconfig import ProcessConfig
from nabu.resources.tasks import build_processing_steps

from nabu.app.logger import Logger

from nabu.app.chunkreader import ChunkReaderComponent
from nabu.app.flatfield import FlatFieldComponent
from nabu.app.phase import PhaseRetrievalComponent
from nabu.app.opmap import NegativeLogComponent
from nabu.app.ccdfilter import CCDFilterComponent
from nabu.app.unsharp import UnsharpMaskComponent
from nabu.app.reconstructor import ReconstructorComponent


if __name__ == "__main__":

    conf = ProcessConfig("/home/pierre/workspace/data/nabu.conf")
    steps, options = build_processing_steps(conf)
    dataset_infos = conf.dataset_infos
    SUB_REGION = (None, None, None, 50)

    logger = Logger("nabu_processing", console=True)

    # Read chunk
    options["read_chunk"]["sub_region"] = SUB_REGION
    options["read_chunk"]["convert_float"] = True
    Ch = ChunkReaderComponent(options["read_chunk"], dataset_infos, logger=logger)
    Ch.execute()
    radios = Ch.chunk_reader.files_data
    d_radios = garray.to_gpu(radios)

    # Flat-field
    options["flatfield"]["sub_region"] = SUB_REGION
    options["flatfield"]["use_opencl"] = False
    options["flatfield"]["use_cuda"] = True
    F = FlatFieldComponent(d_radios, options["flatfield"], dataset_infos, logger=logger)
    F.execute()

    # CCD filter
    options["ccd_correction"]["use_cuda"] = True
    options["ccd_correction"]["use_opencl"] = False
    CCD = CCDFilterComponent(d_radios, options["ccd_correction"], dataset_infos, logger=logger)
    CCD.execute()


    # Phase retrieval
    options["phase"]["use_cuda"] = True
    options["phase"]["use_opencl"] = False
    P = PhaseRetrievalComponent(radios[0].shape, options["phase"], dataset_infos, logger=logger)
    P.execute(d_radios)

    # Unsharp
    # ~ options["unsharp_mask"]["use_cuda"] = True
    # ~ options["unsharp_mask"]["use_opencl"] = False
    # ~ U = UnsharpMaskComponent(radios[0].shape, options["unsharp_mask"], dataset_infos, logger=logger)
    # ~ U.execute(d_radios)

    # -log()
    options["take_log"]["use_cuda"] = True
    options["take_log"]["use_opencl"] = False
    L = NegativeLogComponent(d_radios, options["take_log"], dataset_infos, logger=logger)
    L.execute()


    # Test
    # ~ from spire.utils import ims
    # ~ r = d_radios.get()
    # ~ ims(r[:, 10, :])


    # Reconstruction
    options["reconstruction"]["use_cuda"] = True

    # Test ...
    if 1:
        options["reconstruction"]["start_x"] = 100
        options["reconstruction"]["end_x"] = -100
        options["reconstruction"]["start_y"] = 100
        options["reconstruction"]["end_y"] = -100
        options["reconstruction"]["start_z"] = 0
        options["reconstruction"]["end_z"] = SUB_REGION[-1]-1
        d_rec = garray.zeros((50, 1849, 1849), "f") # y
    #
    else:
        d_rec = garray.zeros((SUB_REGION[-1], 2048, 2048), "f")
    R = ReconstructorComponent(d_radios.shape, options["reconstruction"], dataset_infos, logger=logger)
    R.execute(d_radios, output=d_rec)


    # Write to file
    rec = d_rec.get()
    import numpy as np
    np.save("/home/pierre/tmp/crayon/recs_nabu_vertical.npy", rec)

    """from nabu.reconstruction.reconstructor_cuda import CudaReconstructor
    R = CudaReconstructor(
        radios.shape,
        [0, 1],
        vol_type="projections",
        extra_options={"padding_mode": conf.nabu_config["reconstruction"]["padding_type"]}
    )
    d_recs = garray.zeros((1, 2048, 2048), "f")
    R.reconstruct(d_radios, output=d_recs)
    recs = d_recs.get()
    ims(recs[0], cmap="gray")
    # save
    import numpy as np
    np.save("/tmp/rec0.npy", recs[0])
    """


"""
Notes
  - If the cuda backend is used at step N, it should also be used at step N+1,


otherwise we have to get() the data

"""






data = h5_read(filename)  # radially integrated data
q = hdf5read(filename,'/q')
y = hdf5read(filename,'/y')
omega = hdf5read(filename,'/omega') ;
# y = y(1:ny*nomega)
y = y[:ny*nomega]
# omega = omega(1:ny*nomega) ;
omega = omega[:ny*nomega]


#
# size(X) : height, width, nz
# size(X, 1): height
# flip(X, 2): X[:, i, :] = np.fliplr(X[:, i, :]) for i
#



def get_matlab_to_python_size(arr, dim):
    if arr.ndim == 2:
        # 2D: both Matlab and Python ordering is (Height, Width)
        return arr.shape[dim-1]
    elif arr.ndim == 3:
        # 3D:
        # Matlab: (Height, Width, Nz) ; indices start at 1
        # Python: (Nz, Height, Width) ; indices start at 0
        matlab_to_python = [None, 1, 2, 0]
        dim_py = matlab_to_python[dim]
        if dim_py is None:
            raise ValueError("Matlab index cannot be zero")
        return arr.shape[dim_py]
    else:
        return arr.size


def xrdrec_pyFAI_data(
    ny,
    nomega,
    omega_step,
    acq_method,
    zinger_removal,
    air_mask,
    diode_norm,
    int_norm_roi,
    sample_move_roi,
    sample_thrs,
    c_range,
    vis_rec,
    h5overwrite
):
    """
    FILENAME HDF5 filename containing radially integrated data
    NY number of y steps
    NOMEGA number of omega steps
    OMEGA_STEP angular step
    ACQ_METHOD acquisition methods: FAST_ROT_SLOW_TRANS, FAST_TRANS_STEP_ROT, ZIG_ZAG, FAST_ROT_SLOW_TRANS_OLD
    ZINGER_REMOVAL = 1 if removal performed, 0 otherwise
    AIR_MASK (matrix) mask for air scattering calculation and removal. Set to 1 the pixels used to calculate
             the air scattering contribution and set to 0 he others. SIZE(AIR_MATRIX) = SIZE(Y) = SIZE(OMEGA) = [NY/2 2*NOMEGA]

    DIODE_NORM = 1 if diode normalization is performed, 0 otherwise
    INT_NORM_ROI (string) region of interest for total integrated scattering normalization, empty if no normalization perform
    SAMPLE_MOVE_ROI roi for calculating the center of mass to correct for sample movement (this does also the center of rotation)
    SAMPLE_THRS threshold for calculating the center of mass. SAMPLE_THRS = 0.25 is a typical value
    C_RANGE
        If ACQ_METHOD = ZIG_ZAG
              C_RANGE is a vector containing the center of rotation searching range.
              C_RANGE is expressed in pixel units
        If ACQ_METHOD = FAST_ROT_SLOW_TRANS
              C_RANGE is a scalar defining the rotation axis.
              C_RANGE is expressed in millimiters
        If ACQ_METHOD = FAST_ROT_SLOW_TRANS_OLD
              C_RANGE is a vector containing the center of rotation searching range.
              C_RANGE is expressed in millimiters
    VIS_REC = 1 if visualization of reconstruction performed
    H5OVERWRITE if overwrite previously reconstructed data, 0 if not
    """

    if diode_norm:
        diode = hdf5read(filename,'/diode')
        # s = s ./ diode(:)' * mean(diode) ;
        s = s / diode * np.mean(diode)

    # Here "s" is likely to be 2D, at least for fast_rot_slow_trans**

    if acq_method == 'zig_zag':
        # s = reshape(s,size(s,1),ny,nomega)
        s = s.reshape((nomega, s.shape[-2], ny))
        # s(:,:,2:2:end) = flip(s(:,:,2:2:end),2) ;
        s[1::2, :, :] = np.flip(s[1::2, :, :], axis=0)
        # s = permute(s,[2,3,1]) ;
        s = np.moveaxis(s, [0, 1, 2], [1, 2, 0])
        # figure(1) ; imagesc(sum(s,3)) ; colormap(jet) ; title(sprintf('integrated sinogram %s',strrep(filename,'_','-'))) ;  drawnow ;
        if verbose:
            plt.figure()
            plt.imshow(np.sum(s, axis=0), cmap="jet")
            plt.title("Integrated sinogram")
        # ny = size(s,1) ;
        ny = s.shape[1]
        # nomega = 180 / omega_step
        nomega = 180 / omega_step

    elif acq_method == 'fast_rot_slow_trans':
        # Here s is 2D ?
        # s = s(:,1:ny*nomega);
        s = s[:, :ny*nomega]
        # s = permute(s,[2 1]) ;
        s = np.moveaxis(s, [0, 1], [1, 0])
        # s = reshape(s,2*nomega,ny/2,size(s,2)) ;
        s = s.reshape((s.shape[0], 2*nomega, ny//2))
        # s = permute(s,[2 1 3]) ;
        s = np.moveaxis(s, [0, 1, 2], [1, 0, 2])
        # y = reshape(y,2*nomega,ny/2) ;
        y = y.reshape((2*nomega, ny//2))
        # y = permute(y,[2 1]) ;
        y = np.moveaxis(y, 1, 0)

    elif acq_method == 'fast_rot_slow_trans_old':
        # Here s is 2D ?
        # s = s(:,1:ny*nomega) ;
        s = s[:, :ny*nomega]
        # s = permute(s,[2 1]) ;
        s = np.moveaxis(s, 1, 0)
        # s = reshape(s,2*nomega,ny/2,size(s,2)) ;
        s = s.reshape((s.shape[0], 2*nomega, ny//2))
        # s = permute(s,[2 1 3]) ;
        s = np.moveaxis(s, [0, 1, 2], [1, 0, 2])
        # y = reshape(y,2*nomega,ny/2) ;
        y = y.reshape((2*nomega, ny//2))
        # y = permute(y,[2 1]) ;
        y = np.moveaxis(y, 1, 0)
        # omega = reshape(omega,2*nomega,ny/2) ;
        omega = omega.reshape((2*nomega, ny//2)
        # omega = permute(omega,[2 1]) ;
        omega = np.moveaxis(omega, 1, 0)
    else:
        print("Unknown acquisition method")
        return


    # remove zingers
    if zinger_removal:
        # sn = sino_hot_spot_remove(s,[0:.005:5],[1 5], 1) ;
        sn = sino_hot_spot_remove(s, np.arange(0, 5+0.005, .005), [1, 5], 1)
     else:
         # sn = s ;
         sn = s

    # remove air scattering
    if air_mask is not None:
        ## air_mask seems to be 3D
        ## sum(air_mask) sums along the first dim (height)
        # if sum(sum(air_mask)==0)
        if np.any(air_mask.sum(axis=1) == 0):
            # use the same air scattering pattern for all orientations
            # n = sum(air_mask(:)) ;
            n = air_mask.sum()
            # a = sum( sum( sn .* air_mask(:,:,ones(1,size(sn,3))), 1), 2) / n ;
            o = np.ones(sn.shape[0])
            a = (sn * air_mask[o, :, :]).sum(axis=1, keepdims=True).sum(axis=2, keepdims=True) / n
            # sn = sn - a(ones(1,size(sn,1)),ones(1,size(sn,2)),:) ;
            sn -= a[:, np.ones(sn.shape[1]), np.ones(sn.shape[2])]
        else:
            # use the same specific air scattering pattern for each orientations
            # n = sum(air_mask) ;
            n = air_mask.sum(axis=1, keepdims=True)
            # a = sum( sn .* air_mask(:,:,ones(1,size(sn,3))), 1) ./ n(:,:,ones(1,size(sn,3)));
            o = np.ones(sn.shape[0])
            num = (sn * air_mask[o, :, :]).sum(axis=1, keepdims=True)
            denom = n[o, :, :]
            a = num / denom
            # sn = sn - a(ones(1,size(sn,1)),:,:) ;
            sn -= a[:, np.ones(sn.shape[1]), :]


    # normalize total intensity
    # should do better. For example this is wrong when using highly absorbing samples without radial
    # symmetry. Possible solutions: matlab interp or model the behaviour of sample absorption
    if int_norm_roi is not None:
        # m = eval(sprintf('mean(mean(sn(:,:,%s),3),1)',int_norm_roi)) ; plot(m(:),'o') ; drawnow ;
        # assuming int_norm_roi is a slice object
        m = sn[int_norm_roi, :, :].mean(axis=0, keepdims=True).mean(axis=1, keepdims=True)
        # sn = sn ./ m(ones(1,size(sn,1)),:,ones(1,size(sn,3))) * mean(m(:)) ;
        sn = sn / m[np.ones(sn.shape[0]), np.ones(sn.shape[1]), :] * m.mean()
        # figure(2) ; imagesc(sum(sn,3)) ; colormap(jet) ; title(sprintf('normalized integrated sinogram %s',filename)) ;  drawnow ;
        if verbose:
            plt.figure()
            plt.imshow(sn.sum(axis=0), cmap="jet")
            plt.title("Normalized integrated sinogram")
            plt.show()


    # correct for sample movement using edge detection (automatically center of rotation)
    if sample_move_roi is not None:
        sn = correct_sample_movement(sn, sample_move_roi, omega_step, sample_thrs)
        # rot_cen = (size(sn,1)+1)/2 ;
        rot_cen = (sn.shape[1] - 1)/2 # 0-based indices



    # center of rotation
    if c_range is not None:
       sni, rot_cen = sino_center_of_rotation(sn, acq_method, c_range, nomega, y, omega)
    else:
       sni = sn


    # FBP recosntruction
    v = fbp_reconstruction(sni, nomega, vis_rec)




import numpy as np
import pytest
from nabu.io.utils import get_compacted_dataslices
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from nabu.testutils import compare_arrays, __do_long_tests__, get_file
from nabu.io.reader import ChunkReader


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    cls.dataset_fname = get_file("bamboo_reduced.nx")
    cls.tol = 1e-7


def get_compacted_dataslices_as_sorted_tuples(reader):
    slice_to_tuple = lambda s: (s.start, s.stop, s.step)
    data_slices_tuples = list(
        set(
            [
                slice_to_tuple(u.data_slice())
                for u in get_compacted_dataslices(
                    reader.files, subsampling=reader.dataset_subsampling, begin=reader._files_begin_idx
                ).values()
            ]
        )
    )
    return sorted(data_slices_tuples, key=lambda t: t[0])


@pytest.mark.skipif(not (__do_long_tests__), reason="Use __do_long_tests__ for this test")
@pytest.mark.usefixtures("bootstrap")
class TestChunkReader:
    def test_subsampling(self):
        """
        Test reading data from even/odd projections
        """
        dataset_info = HDF5DatasetAnalyzer(self.dataset_fname)

        # Read all projections, then only even-numbered ones, then odd-numbered ones
        # Use the same object to hopefully use less memory (through garbage collection)
        reader = ChunkReader(dataset_info.projections, dataset_subsampling=None)
        reader.load_data()
        first_sino_all = reader.data[:, 0, :].copy()
        compacted_dataslices_all = get_compacted_dataslices_as_sorted_tuples(reader)

        reader = ChunkReader(dataset_info.projections, dataset_subsampling=(2, 0))
        reader.load_data()
        first_sino_even = reader.data[:, 0, :].copy()
        compacted_dataslices_even = get_compacted_dataslices_as_sorted_tuples(reader)

        reader = ChunkReader(dataset_info.projections, dataset_subsampling=(2, 1))
        reader.load_data()
        first_sino_odd = reader.data[:, 0, :].copy()
        compacted_dataslices_odd = get_compacted_dataslices_as_sorted_tuples(reader)

        # from spire.utils import ims
        # ims([first_sino_all[::2], first_sino_even, first_sino_all[::2]*1. - first_sino_even])
        # ims([first_sino_all[::2], first_sino_even, first_sino_all[1::2]*1. - first_sino_odd])

        # Check that the compacted data slices are correct
        assert len(compacted_dataslices_all) == len(compacted_dataslices_even) == len(compacted_dataslices_odd)
        for data_slice_all, data_slice_even, data_slice_odd in zip(
            compacted_dataslices_all, compacted_dataslices_even, compacted_dataslices_odd
        ):
            indices_all = np.arange(100000)[slice(*data_slice_all)]
            indices_even = np.arange(100000)[slice(*data_slice_even)]
            indices_odd = np.arange(100000)[slice(*data_slice_odd)]
            assert indices_all[::2].size == indices_even.size
            assert np.allclose(indices_all[::2], indices_even)
            assert indices_all[1::2].size == indices_odd.size
            assert np.allclose(indices_all[1::2], indices_odd)

        is_close, max_err = compare_arrays(first_sino_all[::2, :], first_sino_even, self.tol, return_residual=True)
        assert is_close, "ChunkReader: something wrong when subsampling for even projections. max_err=%.3e" % max_err
        is_close, max_err = compare_arrays(first_sino_all[1::2, :], first_sino_odd, self.tol, return_residual=True)
        assert is_close, "ChunkReader: something wrong when subsampling for odd projections. max_err=%.3e" % max_err



import matplotlib.pyplot as plt


def ims(img, cmap=None, legend=None, nocbar=False, share=True):
    """
    image visualization utility.

    img: 2D numpy.ndarray, or list of 2D numpy.ndarray
        image or list of images
    cmap: string
        Optionnal, name of the colorbar to use.
    legend: string, or list of string
        legend under each image
    nocbar: bool
        if True, no colorbar are displayed. Default is False
    share: bool
        if True, the axis are shared between the images, so that zooming in one image
        will zoom in all the corresponding regions of the other images. Default is True
    """
    try:
        _ = img.shape
        nimg = 1
    except AttributeError:
        nimg = len(img)
    #
    if (nimg <= 2): shp = (1,2)
    elif (nimg <= 4): shp = (2,2)
    elif (nimg <= 6): shp = (2,3)
    elif (nimg <= 9): shp = (3,3)
    else: raise ValueError("too many images")
    #
    plt.figure()
    for i in range(nimg):
        curr = list(shp)
        curr.append(i+1)
        curr = tuple(curr)
        if nimg > 1:
            if i == 0: ax0 = plt.subplot(*curr)
            else:
                if share: plt.subplot(*curr, sharex=ax0, sharey=ax0)
                else: plt.subplot(*curr)
            im = img[i]
            if legend: leg = legend[i]
        else:
            im = img
            if legend: leg = legend
        if cmap:
            plt.imshow(im, cmap=cmap, interpolation="nearest")
        else:
            plt.imshow(im, interpolation="nearest")
        if legend: plt.xlabel(leg)
        if nocbar is False: plt.colorbar()

    plt.show()



import numpy as np
from glob import glob
import h5py
from silx.io.dictdump import h5todict

# ~ data = h5todict("/scratch/paleo/bamboo/bamboo_nxtomomill_rec_0100_0.hdf5", path="/entry0000/reconstruction/results")["data"]

def merge_recs(pattern, h5_path):
    fl = glob(pattern)
    fl.sort()
    if fl == []:
        raise ValueError("Nothing found as pattern %s" % pattern)
    recs = []
    n_imgs = 0
    for fname in fl:
        data = h5todict(fname, path=h5_path)["data"]
        n_imgs += data.shape[0]
        recs.append(data)
    recs_np = np.zeros((n_imgs,) + recs[0].shape[1:], dtype=recs[0].dtype)
    pos = 0
    for i in range(len(recs)):
        data = recs[i]
        recs_np[pos:pos+data.shape[0], :, :] = data[:, :, :]
        pos += data.shape[0]
    return recs_np


# TODO checks on shape[1:]
def merge_hdf5_files(pattern, h5_path, output_file):
    files_list = glob(pattern)
    files_list.sort()
    if files_list == []:
        raise ValueError("Nothing found as pattern %s" % pattern)
    virtual_sources = []
    shapes = []
    for fname in files_list:
        with h5py.File(fname, "r") as fid:
            shape = fid[h5_path].shape
        vsource = h5py.VirtualSource(fname, name=h5_path, shape=shape)
        virtual_sources.append(vsource)
        shapes.append(shape)

    n_images = sum([shape[0] for shape in shapes])
    virtual_layout = h5py.VirtualLayout(
        shape=(n_images, ) + shapes[0][1:],
        dtype='f'
    )
    start_idx = 0
    for vsource, shape in zip(virtual_sources, shapes):
        n_imgs = shape[0]
        virtual_layout[start_idx:start_idx + n_imgs] = vsource
        start_idx += n_imgs
    with h5py.File(output_file, 'w') as f:
        f.create_virtual_dataset('data', virtual_layout)





import sys
from os import getpid
from distributed import Client, get_worker


# sur SLURMCluster... peut-on faire autrement ?
def wprint(msg):
    wname = get_worker().name
    fname = "/tmp/workers.log"
    fid = open(fname, "a")
    fid.write(msg + "\n")
    fid.flush()
    fid.close()


def wprint2(msg):
    print(msg, file=sys.stderr, flush=True)




class WorkerProcess:
    def __init__(self, wtype):
        self.wtype = "CPU" if wtype == "CPU" else "GPU"
        self.pid = getpid()
        self._init_worker()

    def _init_worker(self):
        wprint("%s: %s" % (self.wtype, sys.executable))
        if self.wtype == "CPU":
            self._init_cpu()
        else:
            self._init_gpu()
        wprint("[%d] Init %s OK" % (self.pid, self.wtype))

    def _init_cpu(self):
        pass

    def _init_gpu(self):
        wprint("initializing GPU...")
        from nabu.cuda.utils import get_cuda_context, __has_pycuda__, __pycuda_error_msg__
        if __has_pycuda__:
            import pycuda.gpuarray as garray
        if not(__has_pycuda__):
            raise ImportError("Could not import pycuda: %s" % __pycuda_error_msg__)
        self.ctx = get_cuda_context()
        wprint("init GPU OK, context = %s, pycuda = %s - %s" % (str(self.ctx), str(__has_pycuda__), __pycuda_error_msg__))
        # push context ?

    def _process_cpu(self, n):
        return sorted(list(range(n)))

    def _process_gpu(self, n):
        self.ctx.push()
        # ~ wprint("pycuda: %s" % str(__has_pycuda__))
        import pycuda.gpuarray as garray # ??
        d_arr = garray.zeros((2048, 2048), "f")
        for i in range(min(n, 1000)):
            d_arr += 1
        return d_arr.get()

    def process(self, n):
        return {"CPU": self._process_cpu, "GPU": self._process_gpu}[self.wtype](n)


def actor_exec_processing(n, iw=-1):
    w = get_worker()
    worker_process = w.actors[list(w.actors.keys())[iw]]
    wprint("[%d] submit n=%d to %s worker" % (worker_process.pid, n, worker_process.wtype))
    res = worker_process.process(n)
    wprint("ok")
    return res




# ~ client = Client("tcp://160.103.228.120:8786")
# ~ W1 = client.submit(
    # ~ WorkerProcess,
    # ~ "GPU",
    # ~ actor=True,
    # ~ workers=[list(client.has_what().keys())[0]]
# ~ )
# ~ W2 = client.submit(
    # ~ WorkerProcess,
    # ~ "CPU",
    # ~ actor=True,
    # ~ workers=[list(client.has_what().keys())[1]]
# ~ )














from dask_jobqueue import SLURMCluster
S = SLURMCluster(cores=8, processes=2, memory="16GB", project="myproject", walltime="00:15:00", queue="deb9-gpu", extra=['--resources GPU=1'], job_extra=['--gres=gpu:1'], python="/scisoft/users/paleo/.venv/python35/bin/python")
S.scale(1)
client = Client(S)

from time import sleep
sleep(4)

W1 = client.submit(
    WorkerProcess,
    "GPU",
    actor=True,
    workers=[list(client.has_what().keys())[0]]
)
W2 = client.submit(
    WorkerProcess,
    "CPU",
    actor=True,
    workers=[list(client.has_what().keys())[1]]
)



import numpy as np
import astra
from spire.utils import ims
vol = np.load("/scisoft/users/paleo/data/maquette/MRI512cube.npy")

n_x, n_y, n_z = vol.shape
n_a = 1500


def _create_projector_parallel(vol_geom, angles, dx=1., dy=1.):
    n_x, n_z = vol_geom['GridColCount'], vol_geom['GridSliceCount']
    # det_spacing_x, det_spacing_y, det_row_count, det_col_count, angles
    proj_geom = astra.create_proj_geom('parallel3d', dx, dy, n_z, n_x, angles)
    projector, proj_geom = astra.create_projector('cuda3d', proj_geom, vol_geom)
    return projector, proj_geom


def _create_projector_helical(vol_geom, dz, n_angles):
    n_x, n_z = vol_geom['GridColCount'], vol_geom['GridSliceCount']
    vectors = np.zeros((n_angles, 12))
    angles = np.linspace(0, np.pi, n_angles)
    for i in range(n_angles):
        # ray direction
        vectors[i,0] = np.sin(angles[i])
        vectors[i,1] = -np.cos(angles[i])
        vectors[i,2] = 0

        # center of detector
        # ~ vectors[i,3] = 0
        # ~ vectors[i,4] = 0
        vectors[i,5] = i*dz

        # vector from detector pixel (0,0) to (0,1)
        vectors[i,6] = np.cos(angles[i])
        vectors[i,7] = np.sin(angles[i])
        vectors[i,8] = 0;

        # vector from detector pixel (0,0) to (1,0)
        vectors[i,9] = 0
        vectors[i,10] = 0
        vectors[i,11] = 1

    proj_geom = astra.create_proj_geom('parallel3d_vec', n_z, n_x, vectors)
    projector = astra.create_projector('cuda3d', proj_geom, vol_geom)
    return projector, proj_geom


def proj3D(volume, n_a):
    n_x, n_y, n_z = volume.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z) # rows, colums, slices (y, x, z)
    vol_astra = astra.data3d.create('-vol', vol_geom, volume)

    angles = np.linspace(0, np.pi, n_a)
    projector, proj_geom = _create_projector_parallel(vol_geom, angles)

    projs_id, projs = astra.create_sino3d_gpu(vol_astra, proj_geom, vol_geom)
    astra.data3d.clear() #

    return projs


def proj3D_helical(volume, n_a, dz):
    n_x, n_y, n_z = volume.shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z) # rows, colums, slices (y, x, z)
    vol_astra = astra.data3d.create('-vol', vol_geom, volume)

    projector, proj_geom = _create_projector_helical(vol_geom, dz, n_a)

    projs_id, projs = astra.create_sino3d_gpu(vol_astra, proj_geom, vol_geom)
    astra.data3d.clear() #

    return projs


# ~ projs = proj3D_helical(vol, 150, 2)



from .sysutils import *

esrf_environments = {
    "ubuntu20.04_cuda10_ppc64le": {
        "needs_conda": False,
        "base_dir": "/scisoft/tomotools_env/nabu/ubuntu20.04/cuda10.1.243-3/ppc64le"
    },
    "ubuntu20.04_cuda10": {
        "needs_conda": False,
        "base_dir": "/scisoft/tomotools_env/nabu/ubuntu20.04/cuda10.1.243-3/x86_64"
    },
    "deb8_cuda75": {
        "needs_conda": True,
        "base_dir": "/scisoft/tomotools_env/nabu/debian8.11/cuda7.5.18-4~bpo8+1",
    },

}


# Rules for SLURM partitions
slurm_partitions_mappings = {
    "p9gpu": "ubuntu20.04_cuda10_ppc64le",
    "id16a": "ubuntu20.04_cuda10_ppc64le",
}

# Rules for individual nodes
nodes_mappings = {
    "p9-??": "ubuntu20.04_cuda10_ppc64le",
    "gpid16a-180?": "ubuntu20.04_cuda10",
    "gpu2-14??": "deb8_cuda75", # OAR
    "gptomo-nice-0405": "deb8_cuda75",
    "gpid11-nice": "deb8_cuda75",
}

if __name__ == "__main__":
    print(get_env_name(verbose=True))
